from io import BytesIO
from typing import BinaryIO, Callable, Optional, Union

from minimalkv._constants import FOREVER, NOT_SET, VALID_KEY_RE_EXTENDED


class UrlMixin:
    """Mixin to support getting a download URL for keys."""

    _check_valid_key: Callable

    def url_for(self, key: str) -> str:
        """Return a full external URL that can be used to retrieve data at ``key``.

        Only checks whether ``key`` is valid.

        Parameters
        ----------
        key : str
            The key for which to generate a url for.

        Returns
        -------
        url: str
            A string containing a URL to access data at key.

        Raises
        ------
        ValueError
            If the key is not valid.

        """
        self._check_valid_key(key)
        return self._url_for(key)

    def _url_for(self, key: str) -> str:
        """Return a full external URL that can be used to retrieve data at ``key``.

        Parameters
        ----------
        key : str
            The key for which to generate a url for.

        Returns
        -------
        url: str
            A string containing a URL to access data at key.

        """
        raise NotImplementedError


class TimeToLiveMixin:
    """Mixin to allow keys to expire after a certain amount of time.

    This mixin overrides some of the signatures of the api of
    :class:`~minimalkv._key_value_store.KeyValueStore`, albeit in a backwards compatible
    manner.

    Any value given for a time-to-live parameter must be one of the following:

    * A positive ``int`` or ``float``, representing seconds,
    * ``minimalkv._constants.FOREVER``, meaning no expiration
    * ``minimalkv._constants.NOT_SET``, meaning that no TTL configuration will be
      done at all or
    * ``None`` representing the default (see
      :class:`.TimeToLiveMixin`'s ``default_ttl_secs``).

    .. note:: When deriving from :class:`~minimalkv._mixins.TimeToLiveMixin`, the same
       default implementations for ``_put``, ``_put_file`` and
       ``_put_filename`` are provided, except that they all take an additional
       ``ttl_secs`` argument. For more information on how to implement
       backends, see :ref:`implement`.

    """

    ttl_support = True
    """Indicates that a key-value store supports time-to-live features. This
    allows users of stores to test for support using::

      getattr(store, 'ttl_support', False)

    """

    default_ttl_secs = NOT_SET
    _check_valid_key: Callable

    def _valid_ttl(
        self, ttl_secs: Optional[Union[float, int, str]]
    ) -> Union[float, int, str]:
        """Check ``ttl_secs`` for validity and replace ``None`` with default.

        Parameters
        ----------
        ttl_secs: numeric or string or None
            Time to live. Numeric or one of ``minimalkv._constants.FOREVER`` or
            ``minimalkv._constants.NOT_SET``. ``None`` will be replaced with
            ``minimalkv._mixins.TimeToLiveMixin.default_ttl_secs``.

        Raises
        ------
        ValueError:
            If ``ttl_secs`` is neither numeric, ``None`` nor one of ``NOT_SET`` and
            ``FOREVER``.
        ValueError:
            If ``ttl_secs`` is a negative numeric.
        """
        if ttl_secs is None:
            ttl_secs = self.default_ttl_secs

        if ttl_secs in (FOREVER, NOT_SET):
            return ttl_secs

        if not isinstance(ttl_secs, (int, float)):
            raise ValueError("Not a valid ttl_secs value: %r" % ttl_secs)

        if ttl_secs < 0:
            raise ValueError("ttl_secs must not be negative: %r" % ttl_secs)

        return ttl_secs

    def put(
        self, key: str, data: bytes, ttl_secs: Optional[Union[str, float, int]] = None
    ) -> str:
        """Store bytestring data at key.

        If ``ttl_secs`` is a positive number, the key will expire after ``ttl_secs``.
        Other possible values for ``ttl_secs`` are ``minimalkv._constants.FOREVER`` (no expiration)
        and ``minimalkv._constants.NOT_SET`` (no TTL configuration). ``None`` will be replaced with
        ``minimalkv._mixins.TimeToLiveMixin.default_ttl_secs``.

        Parameters
        ----------
        key : str
            The key under which the data is to be stored.
        data : bytes
            Data to be stored at key, must be of types ``bytes``.
        ttl_secs : numeric or str
            Number of seconds until the key expires.

        Returns
        -------
        key: str
            The key under which data was stored.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If storing failed or the file could not be read.
        ValueError
            If ``ttl_secs`` is invalid.

        """
        self._check_valid_key(key)
        if not isinstance(data, bytes):
            raise OSError("Provided data is not of type bytes")
        return self._put(key, data, self._valid_ttl(ttl_secs))

    def put_file(
        self,
        key: str,
        file: Union[str, BinaryIO],
        ttl_secs: Optional[Union[float, int, str]] = None,
    ) -> str:
        """Store contents of file at key.

        Store data from a file into key. ``file`` can be a string, which will be
        interpreted as a filename, or an object with a ``read()`` method.

        If ``file`` is a filename, the file might be removed while storing to avoid
        unnecessary copies. To prevent this, pass the opened file instead.

        If ``ttl_secs`` is a positive number, the key will expire after ``ttl_secs``.
        Other possible values for ``ttl_secs`` are `minimalkv._constants.FOREVER` (no expiration)
        and ``minimalkv._constants.NOT_SET`` (no TTL configuration). ``None`` will be replaced with
        ``minimalkv._mixins.TimeToLiveMixin.default_ttl_secs``.

        Parameters
        ----------
        key : str
            Key where to store data in file.
        file : BinaryIO or str
            A filename or an object with a read method.
        ttl_secs : str or numeric or None, optional, default = None
            Number of seconds until the key expires.

        Returns
        -------
        key: str
            The key under which data was stored.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If there was a problem moving the file in.
        ValueError
            If ``ttl_secs`` is invalid.

        """
        self._check_valid_key(key)

        if isinstance(file, str):
            return self._put_filename(key, file, self._valid_ttl(ttl_secs))
        else:
            return self._put_file(key, file, self._valid_ttl(ttl_secs))

    # default implementations similar to KeyValueStore below:
    def _put(
        self, key: str, data: bytes, ttl_secs: Optional[Union[str, float, int]] = None
    ) -> str:
        """Store bytestring data at key.

        Parameters
        ----------
        key : str
            Key under which data should be stored.
        data : bytes
            Data to be stored.
        ttl_secs : str or numeric or None, optional, default = None
            Number of seconds until the key expires.

        Returns
        -------
        key : str
            Key where data was stored.

        """
        return self._put_file(key, BytesIO(data), ttl_secs)

    def _put_file(
        self,
        key: str,
        file: BinaryIO,
        ttl_secs: Optional[Union[str, float, int]] = None,
    ):
        """Store contents of file at key.

        Parameters
        ----------
        key : str
            Key under which data should be stored.
        file : BinaryIO
            File-like object with a ``read`` method.
        ttl_secs : str or numeric or None, optional, default = None
            Number of seconds until the key expires.

        Returns
        -------
        key : str
            Key where data was stored.

        """
        raise NotImplementedError

    def _put_filename(
        self, key: str, filename: str, ttl_secs: Optional[Union[str, float, int]] = None
    ):
        """Store contents of file at key.

        Parameters
        ----------
        key : str
            Key under which data should be stored.
        filename : str
            Filename of file from which to read data.
        ttl_secs : str or numeric or None, optional, default = None
            Number of seconds until the key expires.

        Returns
        -------
        key : str
            Key where data was stored.

        """
        with open(filename, "rb") as source:
            return self._put_file(key, source, self._valid_ttl(ttl_secs))


class CopyMixin:
    """Mixin to expose a copy operation supported by the backend."""

    _check_valid_key: Callable
    _delete: Callable

    def copy(self, source: str, dest: str) -> str:
        """Copy data at key ``source`` to key ``dest``.

        The destination is overwritten if it does exist.

        Parameters
        ----------
        source : str
            The source key of data to copy.
        dest : str
            The destination for the copy.

        Raises
        ------
        ValueError
            If ``source`` is not a valid key.
        ValueError
            If ``dest`` is not a valid key.

        Returns
        -------
        key : str
            The destination key.

        """
        self._check_valid_key(source)
        self._check_valid_key(dest)
        return self._copy(source, dest)

    def _copy(self, source: str, dest: str):
        """Copy data at key ``source`` to key ``dest``.

        The destination is overwritten if it does exist.

        Parameters
        ----------
        source : str
            The source key of data to copy.
        dest : str
            The destination for the copy.

        Returns
        -------
        key : str
            The destination key.

        """
        raise NotImplementedError

    def move(self, source: str, dest: str) -> str:
        """Move data from key ``source`` to key ``dest``.

        The destination is overwritten if it does exist.

        Parameters
        ----------
        source : str
            The source key of data to be moved.
        dest : str
            The destination for the data.

        Raises
        ------
        ValueError
            If ``source`` is not a valid key.
        ValueError
            If ``dest`` is not a valid key.

        Returns
        -------
        key : str
            The destination key.

        """
        self._check_valid_key(source)
        self._check_valid_key(dest)
        return self._move(source, dest)

    def _move(self, source: str, dest: str) -> str:
        """Move data from key ``source`` to key ``dest``.

        The destination is overwritten if it does exist.

        Parameters
        ----------
        source : str
            The source key of data to be moved.
        dest : str
            The destination for the data.

        Returns
        -------
        key : str
            The destination key.

        """
        self._copy(source, dest)
        self._delete(source)
        return dest


class ExtendedKeyspaceMixin:
    """A mixin to extend the keyspace to allow slashes and spaces in keynames.

    Attention: A single / is NOT allowed.
    Use it by extending first from ` :class:`~minimalkv._mixins.ExtendedKeyspaceMixin`
    and then by the desired store.
    Note: This Mixin is unsupported and might not work correctly with all backends.

    """

    def _check_valid_key(self, key: Optional[str]) -> None:
        """Check if a key is valid and raises a ValueError if its not.

        When in need of checking a key for validity, always use this
        method if possible.

        Parameters
        ----------
        key : str
            The key to be checked

        """
        if key is not None:
            if not isinstance(key, str):
                raise ValueError("%r is not a valid key type" % key)
            elif not VALID_KEY_RE_EXTENDED.match(key) or key == "/":
                raise ValueError("%r contains illegal characters" % key)



import io
from contextlib import contextmanager
from shutil import copyfileobj
from typing import Iterator, List, Optional

from minimalkv import CopyMixin, KeyValueStore, UrlMixin


def _public_readable(grants: List) -> bool:  # TODO: What kind of list
    """Take a list of grants from an ACL and check if they allow public read access."""
    for grant in grants:
        # see: https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html
        if grant["Permission"] not in ("READ", "FULL_CONTROL"):
            continue
        grantee = grant["Grantee"]
        if grantee.get("Type") != "Group":
            continue
        if grantee.get("URI") != "http://acs.amazonaws.com/groups/global/AllUsers":
            continue
        return True
    return False


@contextmanager
def map_boto3_exceptions(key=None, exc_pass=()):
    """Map boto3-specific exceptions to the minimalkv-API."""
    from botocore.exceptions import ClientError

    try:
        yield
    except ClientError as ex:
        code = ex.response["Error"]["Code"]
        if code == "404" or code == "NoSuchKey":
            raise KeyError(key) from ex
        raise OSError(str(ex)) from ex


class Boto3SimpleKeyFile(io.RawIOBase):  # noqa D
    # see: https://alexwlchan.net/2019/02/working-with-large-s3-objects/
    # author: Alex Chan, license: MIT
    def __init__(self, s3_object):
        self.s3_object = s3_object
        self.position = 0

    def __repr__(self):  # noqa D
        return f"<{type(self).__name__} s3_object={self.s3_object!r} >"

    @property
    def size(self) -> int:  # noqa D
        return self.s3_object.content_length

    def tell(self) -> int:  # noqa D
        return self.position

    def seek(self, offset: int, whence=io.SEEK_SET) -> int:  # noqa D
        if whence == io.SEEK_SET:
            self.position = offset
        elif whence == io.SEEK_CUR:
            self.position += offset
        elif whence == io.SEEK_END:
            self.position = self.size + offset
        else:
            raise ValueError(
                "invalid whence (%r, should be %d, %d, %d)"
                % (whence, io.SEEK_SET, io.SEEK_CUR, io.SEEK_END)
            )

        return self.position

    def seekable(self) -> bool:  # noqa D
        return True

    def read(self, size=-1):  # noqa D
        if size == -1:
            # Read to the end of the file
            range_header = "bytes=%d-" % self.position
            self.seek(offset=0, whence=io.SEEK_END)
        else:
            new_position = self.position + size

            # If we're going to read beyond the end of the object, return
            # the entire object.
            if new_position >= self.size:
                return self.read()

            range_header = "bytes=%d-%d" % (self.position, new_position - 1)
            self.seek(offset=size, whence=io.SEEK_CUR)

        return self.s3_object.get(Range=range_header)["Body"].read()

    def readable(self) -> bool:  # noqa D
        return True


class Boto3Store(KeyValueStore, UrlMixin, CopyMixin):  # noqa D
    def __init__(
        self,
        bucket,
        prefix: Optional[str] = None,
        object_prefix: str = "",
        url_valid_time: int = 0,
        reduced_redundancy: bool = False,
        public: bool = False,
        metadata=None,
        create_if_missing=False,
    ):
        import boto3

        if isinstance(bucket, str):
            s3_resource = boto3.resource("s3")
            bucket = s3_resource.Bucket(bucket)
            if bucket not in s3_resource.buckets.all():
                raise ValueError("invalid s3 bucket name")
        self.bucket = bucket

        if prefix is not None:
            import warnings

            warnings.warn(
                "The prefix attribute is deprecated and will be removed in the next major release."
                "Use object_prefix instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            object_prefix = object_prefix or prefix
        self._object_prefix = object_prefix.strip().lstrip("/")

        self.url_valid_time = url_valid_time
        self.reduced_redundancy = reduced_redundancy
        self.public = public
        self.metadata = metadata or {}

    @property
    def prefix(self) -> str:
        """Get the prefix used for all keys in this store.

        .. note:: Deprecated in 2.0.0, use :attr:`object_prefix` instead.
        """
        import warnings

        warnings.warn(
            "The `prefix` attribute is deprecated and will be removed in the next major release."
            "Use `object_prefix` instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self._object_prefix

    def __new_object(self, name):
        return self.bucket.Object(self._object_prefix + name)

    def iter_keys(self, prefix: str = "") -> Iterator[str]:  # noqa D
        with map_boto3_exceptions():
            prefix_len = len(self._object_prefix)
            return (
                k.key[prefix_len:]
                for k in self.bucket.objects.filter(Prefix=self._object_prefix + prefix)
            )

    def _delete(self, key: str) -> None:
        self.bucket.Object(self._object_prefix + key).delete()

    def _get(self, key: str) -> bytes:
        obj = self.__new_object(key)
        with map_boto3_exceptions(key=key):
            obj = obj.get()
            return obj["Body"].read()

    def _get_file(self, key: str, file):
        obj = self.__new_object(key)
        with map_boto3_exceptions(key=key):
            obj = obj.get()
            return copyfileobj(obj["Body"], file)

    def _get_filename(self, key: str, filename):
        obj = self.__new_object(key)
        with map_boto3_exceptions(key=key):
            obj = obj.get()
            with open(filename, "wb") as file:
                return copyfileobj(obj["Body"], file)

    def _open(self, key: str):
        obj = self.__new_object(key)
        with map_boto3_exceptions(key=key):
            obj.load()
            return Boto3SimpleKeyFile(obj)

    def _copy(self, source: str, dest: str) -> None:
        obj = self.__new_object(dest)
        parameters = {
            "CopySource": self.bucket.name + "/" + self._object_prefix + source,
            "Metadata": self.metadata,
        }
        if self.public:
            parameters["ACL"] = "public-read"
        if self.reduced_redundancy:
            parameters["StorageClass"] = "REDUCED_REDUNDANCY"
        with map_boto3_exceptions(key=source):
            self.__new_object(source).load()
            obj.copy_from(**parameters)

    def _put(self, key: str, data) -> str:
        obj = self.__new_object(key)
        parameters = {"Body": data, "Metadata": self.metadata}
        if self.public:
            parameters["ACL"] = "public-read"
        if self.reduced_redundancy:
            parameters["StorageClass"] = "REDUCED_REDUNDANCY"
        with map_boto3_exceptions(key=key):
            obj.put(**parameters)
        return key

    def _put_file(self, key: str, file) -> str:
        return self._put(key, file)

    def _put_filename(self, key: str, filename) -> str:
        with open(filename, "rb") as file:
            return self._put(key, file)

    def _url_for(self, key: str) -> str:
        import boto3
        import botocore.client
        import botocore.exceptions

        obj = self.__new_object(key)
        try:
            grants = obj.Acl().grants
        except botocore.exceptions.ClientError:
            is_public = False
        else:
            is_public = _public_readable(grants)
        if self.url_valid_time and not is_public:
            s3_client = boto3.client(
                "s3", endpoint_url=self.bucket.meta.client.meta.endpoint_url
            )
        else:
            s3_client = boto3.client(
                "s3",
                config=botocore.client.Config(signature_version=botocore.UNSIGNED),
                endpoint_url=self.bucket.meta.client.meta.endpoint_url,
            )
        with map_boto3_exceptions(key=key):
            return s3_client.generate_presigned_url(
                "get_object",
                Params={"Bucket": self.bucket.name, "Key": key},
                ExpiresIn=self.url_valid_time,
            )

    def __eq__(self, other):
        """Assert that two ``Boto3Store``s are equal.

        The bucket name and other configuration parameters are compared.
        See :func:`from_url` for details on the connection parameters.
        Does NOT compare the credentials or the contents of the bucket!
        """
        return (
            isinstance(other, Boto3Store)
            and self.bucket.name == other.bucket.name
            and self.bucket.meta.client.meta.endpoint_url
            == other.bucket.meta.client.meta.endpoint_url
            and self._object_prefix == other._object_prefix
            and self.url_valid_time == other.url_valid_time
            and self.reduced_redundancy == other.reduced_redundancy
            and self.public == other.public
            and self.metadata == other.metadata
        )



from configparser import ConfigParser
from contextlib import contextmanager
from uuid import uuid4 as uuid

import boto3
import pytest

boto = pytest.importorskip("boto")


@contextmanager
def boto_bucket(
    access_key,
    secret_key,
    host,
    connect_func="connect_s3",
    ordinary_calling_format=False,
    bucket_name=None,
    port=None,
    is_secure=True,
):
    if ordinary_calling_format:
        from boto.s3.connection import OrdinaryCallingFormat

        conn = getattr(boto, connect_func)(
            access_key,
            secret_key,
            host=host,
            calling_format=OrdinaryCallingFormat(),
            port=port,
            is_secure=is_secure,
        )
    else:
        conn = getattr(boto, connect_func)(
            access_key, secret_key, host=host, port=port, is_secure=is_secure
        )

    name = bucket_name or f"testrun-bucket-{uuid()}"
    bucket = conn.create_bucket(name)

    yield bucket

    for key in bucket.list():
        key.delete()
    bucket.delete()


@contextmanager
def boto3_bucket(
    access_key,
    secret_key,
    host=None,
    bucket_name=None,
    port=None,
    is_secure=None,
    **kwargs,
):
    """Create a boto3 bucket.

    The bucket is deleted after the consuming function returns.
    """
    bucket = boto3_bucket_reference(
        host=host,
        bucket_name=bucket_name,
        port=port,
        is_secure=is_secure,
        access_key=access_key,
        secret_key=secret_key,
    )
    bucket.create()

    yield bucket

    for key in bucket.objects.all():
        key.delete()
    bucket.delete()


def boto3_bucket_reference(
    access_key=None,
    secret_key=None,
    host=None,
    bucket_name=None,
    port=None,
    is_secure=None,
):
    """Create a boto3 bucket reference.

    The bucket is not created.
    """
    import os

    # Set environment variables for boto3
    os.environ["AWS_ACCESS_KEY_ID"] = access_key
    os.environ["AWS_SECRET_ACCESS_KEY"] = secret_key
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"

    # Build endpoint host
    endpoint_url = None
    if host:
        scheme = "https" if is_secure else "http"
        endpoint_url = f"{scheme}://{host}"
        if port:
            endpoint_url += f":{port}"

    name = bucket_name or f"testrun-bucket-{uuid()}"
    # We only set the endpoint url if we're testing against a non-aws host
    if port != 80:
        s3_resource = boto3.resource(
            "s3",
            endpoint_url=endpoint_url,
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            region_name="us-east-1",
        )
    else:
        s3_resource = boto3.resource(
            "s3",
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            region_name="us-east-1",
        )

    bucket = s3_resource.Bucket(name)
    return bucket


def load_boto_credentials():
    # loaded from the same place tox.ini. here's a sample
    #
    # [my-s3]
    # access_key=foo
    # secret_key=bar
    # connect_func=connect_s3
    #
    # [my-gs]
    # access_key=foo
    # secret_key=bar
    # connect_func=connect_gs
    cfg_fn = "boto_credentials.ini"

    parser = ConfigParser(
        {
            "host": "s3.amazonaws.com",
            "is_secure": "true",
            "ordinary_calling_format": "false",
        }
    )
    if not parser.read(cfg_fn):
        pytest.skip(f"file {cfg_fn} not found")

    for section in parser.sections():
        yield {
            "access_key": parser.get(section, "access_key"),
            "secret_key": parser.get(section, "secret_key"),
            "connect_func": parser.get(section, "connect_func"),
            "host": parser.get(section, "host"),
            "is_secure": parser.getboolean(section, "is_secure"),
            "port": parser.getint(section, "port"),
            "ordinary_calling_format": parser.getboolean(
                section, "ordinary_calling_format"
            ),
        }


boto_credentials = list(load_boto_credentials())



import io
import warnings
from typing import TYPE_CHECKING, BinaryIO, Iterator, Optional, Union

from minimalkv.net._net_common import LAZY_PROPERTY_ATTR_PREFIX, lazy_property

if TYPE_CHECKING:
    from fsspec import AbstractFileSystem
    from fsspec.spec import AbstractBufferedFile

from minimalkv import KeyValueStore

# The complete path of the key is structured as follows:
# /Users/simon/data/mykvstore/file1
# <prefix>                    <key>
# If desired to be a directory, the prefix should end in a slash.


class FSSpecStoreEntry(io.BufferedIOBase):
    """A file-like object for reading from an entry in an FSSpecStore."""

    def __init__(self, file: "AbstractBufferedFile"):
        """Initialize an FSSpecStoreEntry.

        Parameters
        ----------
        file: AbstractBufferedFile
            The fsspec file object to wrap.
        """
        super().__init__()
        self._file = file

    def seek(self, loc: int, whence: int = 0) -> int:
        """Set current file location.

        Parameters
        ----------
        loc: int
            byte location
        whence: {0, 1, 2}
            from start of file, current location or end of file, respectively.
        """
        if self.closed:
            raise ValueError("I/O operation on closed file.")
        try:
            return self._file.seek(loc, whence)
        except ValueError as e:
            # Map ValueError to IOError
            raise OSError from e

    def tell(self) -> int:
        """Return the current offset as int. Always >= 0."""
        if self.closed:
            raise ValueError("I/O operation on closed file.")
        return self._file.tell()

    def read(self, size: Optional[int] = -1) -> bytes:
        """Return first ``size`` bytes of data.

        If no size is given all data is returned.

        Parameters
        ----------
        size : int, optional, default = -1
            Number of bytes to be returned.

        """
        if self.closed:
            raise ValueError("I/O operation on closed file")
        return self._file.read(size)

    def seekable(self) -> bool:
        """Whether the file is seekable."""
        return self._file.seekable()

    def readable(self) -> bool:
        """Whether the file is readable."""
        return self._file.readable()


class FSSpecStore(KeyValueStore):
    """A KeyValueStore that uses an fsspec AbstractFileSystem to store the key-value pairs."""

    def __init__(
        self,
        prefix: str = "",
        mkdir_prefix: bool = True,
        write_kwargs: Optional[dict] = None,
        custom_fs: Optional["AbstractFileSystem"] = None,
    ):
        """Initialize an FSSpecStore.

        The underlying fsspec FileSystem is created when the store is used for the first time.

        Parameters
        ----------
        prefix: str, optional
            The prefix to use on the FSSpecStore when storing keys.
        mkdir_prefix: bool, optional
            If True, the prefix will be created if it does not exist.
            Analogous to the create_if_missing parameter in AzureBlockBlobStore or GoogleCloudStore.
        write_kwargs: dict, optional
            Additional keyword arguments to pass to the fsspec FileSystem when writing files.
        custom_fs: AbstractFileSystem, optional
            If given, use this fsspec FileSystem instead of creating a new one.
        """
        write_kwargs = write_kwargs or {}
        self._prefix = prefix
        self._mkdir_prefix = mkdir_prefix
        self._write_kwargs = write_kwargs
        self._custom_fs = custom_fs

    @lazy_property
    def _prefix_exists(self) -> Union[None, bool]:
        from google.auth.exceptions import RefreshError

        # Check if prefix exists.
        # Used by inheriting classes to check if e.g. a bucket exists.
        try:
            return self._fs.exists(self._prefix)
        except (OSError, RefreshError):
            return None

    @property
    def mkdir_prefix(self):
        """Whether to create the prefix if it does not exist.

        .. note:: Deprecated in 2.0.0.
        """
        warnings.warn(
            "The mkdir_prefix attribute is deprecated!"
            "It will be renamed to _mkdir_prefix in the next release.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self._mkdir_prefix

    @property
    def prefix(self):
        """Get the prefix used on the ``fsspec`` ``FileSystem`` when storing keys.

        .. note:: Deprecated in 2.0.0.
        """
        warnings.warn(
            "The prefix attribute is deprecated!"
            "It will be renamed to _prefix in the next release.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self._prefix

    def iter_keys(self, prefix: str = "") -> Iterator[str]:
        """Iterate over all keys in the store starting with prefix.

        Parameters
        ----------
        prefix: str, optional, default = ''
            Only iterate over keys starting with prefix. Iterate over all keys if empty.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        # We want to look for files whose path starts with the full prefix.
        # `find` lists all files in a directory and allows
        # limiting results to files whose names start with a given prefix.
        # Thus, we have to split the full path into a directory and a file prefix.
        full_prefix = f"{self._prefix}{prefix}"
        # Find last slash in full prefix
        last_slash = full_prefix.rfind("/")
        if last_slash == -1:
            # No slash in full prefix
            dir_prefix = ""
            file_prefix = full_prefix
        else:
            dir_prefix = full_prefix[: last_slash + 1]
            file_prefix = full_prefix[last_slash + 1 :]

        all_files_and_dirs = self._fs.find(dir_prefix, prefix=file_prefix)

        return (k.replace(f"{self._prefix}", "") for k in all_files_and_dirs)

    def _delete(self, key: str) -> None:
        try:
            self._fs.rm_file(f"{self._prefix}{key}")
        except FileNotFoundError:
            pass

    def _open(self, key: str) -> BinaryIO:
        try:
            return self._fs.open(f"{self._prefix}{key}")
        except FileNotFoundError as e:
            raise KeyError(key) from e

    # Required to prevent error when credentials are not sufficient for listing objects
    def _get_file(self, key: str, file: BinaryIO) -> str:
        try:
            file.write(self._fs.cat_file(f"{self._prefix}{key}"))
            return key
        except FileNotFoundError as e:
            raise KeyError(key) from e

    def _put_file(self, key: str, file: BinaryIO) -> str:
        self._fs.pipe_file(f"{self._prefix}{key}", file.read(), **self._write_kwargs)
        return key

    def _has_key(self, key: str) -> bool:
        return self._fs.exists(f"{self._prefix}{key}")

    def _create_filesystem(self) -> "AbstractFileSystem":
        # To be implemented by inheriting classes.
        raise NotImplementedError

    @lazy_property
    def _fs(self) -> "AbstractFileSystem":
        if self._custom_fs is not None:
            fs = self._custom_fs
        else:
            fs = self._create_filesystem()

        if self._mkdir_prefix and not fs.exists(self._prefix):
            fs.mkdir(self._prefix)
        return fs

    # Skips lazy properties.
    # These will be recreated after unpickling through the lazy_property decorator
    def __getstate__(self) -> dict:  # noqa D
        return {
            key: value
            for key, value in self.__dict__.items()
            if not key.startswith(LAZY_PROPERTY_ATTR_PREFIX)
        }



import hashlib

import pytest


@pytest.fixture(params=["sha1", "sha256", "md5"])
def hashfunc(request):
    return getattr(hashlib, request.param)


@pytest.fixture(params=[b"secret_key_a", b"\x12\x00\x12test"])
def secret_key(request):
    return request.param


# values are short payloads to store
@pytest.fixture(params=[b"a_short_value"])
def value(request):
    return request.param


@pytest.fixture(params=[b"the_other_value"])
def value2(request):
    return request.param


@pytest.fixture(params=["the_other_value", "othäöü_valਊਏਐਓਔਕਖਗue_2"])
def unicode_value(request):
    return request.param


@pytest.fixture(params=[b"a_long_value" * 4 * 1024])
def long_value(request):
    return request.param


# keys are always strings. only ascii chars are allowed
@pytest.fixture(params=["short_key", """'!"`#$%&'()+,-.<=>?@[]^_{}~'"""])
def key(request):
    return request.param


@pytest.fixture(params=["key_number_2"])
def key2(request):
    return request.param


@pytest.fixture(params=["ä", "/", "\x00", "*", "", "no whitespace allowed"])
def invalid_key(request):
    return request.param


@pytest.fixture(params=[b"short_key", b"""'!"`#$%&'()+,-.<=>?@[]^_{}~'"""])
def bytestring_key(request):
    return request.param


# maximum length key
# 230 is chosen because the fake-gcs-server stores the objects in the filesystem
# and appends a hash for uniqueness. The hash is roughly 15 bytes long, and the maximum filename length is 255.
# The maximum object name length for live S3 and GCS is 1024 bytes.
@pytest.fixture(params=["a" * 230])
def max_key(request):
    return request.param


# Test class to derive from to get test fixtures for the extended keyspace
class ExtendedKeyspaceTests:
    @pytest.fixture(
        params=[
            "short ke/y",
            "short_key",
            """'!"`#$%&'()+,-.<=>?@[]^_{}~/'""",
        ]
    )
    def key(self, request):
        return request.param

    @pytest.fixture(params=["key_number/2 with space"])
    def key2(self, request):
        return request.param

    @pytest.fixture(params=["ä", "/", "\x00", "*", ""])
    def invalid_key(self, request):
        return request.param



import base64
from typing import Any, Dict, List

from uritools import urisplit

TRUEVALUES = ("true",)


def url2dict(url: str, raise_on_extra_params: bool = False) -> Dict[str, Any]:
    """Create dictionary with parameters from url.

    Parameters
    ----------
    url : str
        Access-URL, see below for supported forms.
    raise_on_extra_params : bool, optional, default = False
        Whether to raise on unexpected params.

    Returns
    -------
    params : dict
        Parameter dictionary suitable for get_store()

    Note
    ----
    Supported formats:
        ``memory://``
        ``redis://[[password@]host[:port]][/db]``
        ``fs://path``
        ``s3://access_key:secret_key@endpoint/bucket[?create_if_missing=true]``
        ``azure://account_name:account_key@container[?create_if_missing=true][?max_connections=2]``
        ``azure://account_name:shared_access_signature@container?use_sas&create_if_missing=false[?max_connections=2&socket_timeout=(20,100)]``
        ``azure://account_name:shared_access_signature@container?use_sas&create_if_missing=false[?max_connections=2&socket_timeout=(20,100)][?max_block_size=4*1024*1024&max_single_put_size=64*1024*1024]``
        ``gcs://<base64 encoded credentialsJSON>@bucket_name[?create_if_missing=true][?bucket_creation_location=EUROPE-WEST1]``
    """
    u = urisplit(url)
    parsed = {
        "scheme": u.getscheme(),
        "host": u.gethost(),
        "port": u.getport(),
        "path": u.getpath(),
        "query": u.getquerydict(),
        "userinfo": u.getuserinfo(),
    }
    fragment = u.getfragment()

    params = {"type": parsed["scheme"]}

    # handling special instructions embedded in the 'fragment' part of the URL,
    # currently only wrappers/store decorators
    fragments = fragment.split("#") if fragment else []
    wrap_spec = list(filter(lambda s: s.startswith("wrap:"), fragments))
    if wrap_spec:
        wrappers = wrap_spec[-1].partition("wrap:")[2]  # remove the 'wrap:' part
        params["wrap"] = wrappers

    if "create_if_missing" in parsed["query"]:
        create_if_missing = parsed["query"].pop("create_if_missing")[
            -1
        ]  # use last appearance of key
        params["create_if_missing"] = create_if_missing in TRUEVALUES

    # get store-specific parameters:
    store_params = extract_params(**parsed)
    params.update(store_params)
    return params


def extract_params(scheme, host, port, path, query, userinfo):  # noqa D
    # We want to ignore wrappers here
    store_type = scheme.split("+")[0]

    if store_type in ("memory", "hmemory"):
        return {}
    if store_type in ("redis", "hredis"):
        path = path[1:] if path.startswith("/") else path
        params = {"host": host or "localhost"}
        if port:
            params["port"] = port
        if userinfo:
            params["password"] = userinfo
        if path:
            params["db"] = int(path)
        return params
    if store_type in ("gcs", "hgcs"):
        credentials_b64 = userinfo
        params = {"type": store_type, "bucket_name": host}
        params["credentials"] = base64.urlsafe_b64decode(credentials_b64.encode())
        if "bucket_creation_location" in query:
            params["bucket_creation_location"] = query.pop("bucket_creation_location")[
                0
            ]
        return params
    if store_type in ("fs", "hfs"):
        return {"type": store_type, "path": host + path}
    if store_type in ("s3", "hs3"):
        access_key, secret_key = _parse_userinfo(userinfo)
        params = {
            "host": f"{host}:{port}" if port else host,
            "access_key": access_key,
            "secret_key": secret_key,
            "bucket": path[1:],
        }
        return params
    if store_type in ("azure", "hazure"):
        account_name, account_key = _parse_userinfo(userinfo)
        params = {
            "account_name": account_name,
            "account_key": account_key,
            "container": host,
        }
        if "use_sas" in query:
            params["use_sas"] = True
        if "max_connections" in query:
            params["max_connections"] = int(query.pop("max_connections")[-1])
        if "socket_timeout" in query:
            params["socket_timeout"] = query.pop("socket_timeout")
        if "max_block_size" in query:
            params["max_block_size"] = query.pop("max_block_size")
        if "max_single_put_size" in query:
            params["max_single_put_size"] = query.pop("max_single_put_size")
        return params

    raise ValueError(f'Unknown storage type "{store_type}"')


def _parse_userinfo(userinfo: str) -> List[str]:
    """Try to split the URL's userinfo into fields separated by `:`.

    The user info is the part between ``://`` and ``@``. If anything looks wrong, remind
    the user to percent-encode values.

    Parameters
    ----------
    userinfo : str
        URL-encoded user-info.

    Returns
    -------
    parts: list of str
        URL-encoded user-info split at ``:``.
    """
    if hasattr(userinfo, "split"):
        parts = userinfo.split(":", 1)

        if len(parts) == 2:
            return parts

    raise ValueError(
        "Could not parse user/key in store-URL. Note that values have to be "
        "percent-encoded, eg. with urllib.quote_plus()."
    )



#!/usr/bin/env python

import pytest
from basic_store import BasicStore, TTLStore
from conftest import ExtendedKeyspaceTests

from minimalkv._mixins import ExtendedKeyspaceMixin

redis = pytest.importorskip("redis")

from redis import StrictRedis
from redis.exceptions import ConnectionError


class TestRedisStore(TTLStore, BasicStore):
    @pytest.fixture
    def store(self):
        from minimalkv.memory.redisstore import RedisStore

        r = StrictRedis()

        try:
            r.get("anything")
        except ConnectionError:
            pytest.skip("Could not connect to redis server")

        r.flushdb()
        with RedisStore(r) as store:
            yield store
        r.flushdb()


class TestExtendedKeyspaceDictStore(TestRedisStore, ExtendedKeyspaceTests):
    @pytest.fixture
    def store(self):
        from minimalkv.memory.redisstore import RedisStore

        class ExtendedKeyspaceStore(ExtendedKeyspaceMixin, RedisStore):
            pass

        r = StrictRedis()

        try:
            r.get("anything")
        except ConnectionError:
            pytest.skip("Could not connect to redis server")

        r.flushdb()
        with ExtendedKeyspaceStore(r) as store:
            yield store
        r.flushdb()



from typing import BinaryIO, Union

from minimalkv._key_value_store import KeyValueStore
from minimalkv.decorator import StoreDecorator


class CacheDecorator(StoreDecorator):
    """Write-through cache decorator.

    Can combine two :class:`~minimalkv._key_value_store.KeyValueStore` instances into a
    single caching :class:`~minimalkv._key_value_store.KeyValueStore`. On a data-read
    request, the cache will be consulted first, if there is a cache miss or an error,
    data will be read from the backing store. Any retrieved value will be stored in the
    cache before being forward to the client.

    Write, key-iteration and delete requests will be passed on straight to the
    backing store. After their completion, the cache will be updated.

    No cache maintainenace above this is done by the decorator. The caching
    store itselfs decides how large to grow the cache and which data to keep,
    which data to throw away.

    Parameters
    ----------
    cache : KeyValueStore
        The caching backend.
    store : KeyValueStore
        The backing store. This is the "authorative" backend.
    """

    def __init__(self, cache: KeyValueStore, store: KeyValueStore):
        super().__init__(store)
        self.cache = cache

    def delete(self, key: str) -> None:
        """Delete data at key.

        Deletes data from both the cache and the backing store.

        Parameters
        ----------
        key : str
            Key of data to be deleted.
        """
        self._dstore.delete(key)
        self.cache.delete(key)

    def get(self, key: str) -> bytes:
        """Return data at key as a bytestring.

        If a cache miss occurs, the value is retrieved, stored in the cache and
        returned.

        If the cache raises an :exc:`~IOError`, the cache is ignored, and the backing
        store is consulted directly.

        It is possible for a caching error to occur while attempting to store the value
        in the cache. It will not be handled as well.

        Parameters
        ----------
        key : str
            The key to be read.

        Returns
        -------
        data : bytes
            Value associated with the key as a ``bytes`` object.

        """
        try:
            return self.cache.get(key)
        except KeyError:
            # cache miss or error, retrieve from backend
            data = self._dstore.get(key)

            # store in cache and return
            self.cache.put(key, data)
            return data
        except OSError:
            # cache error, ignore completely and return from backend
            return self._dstore.get(key)

    def get_file(self, key: str, file: Union[str, BinaryIO]) -> str:
        """Write data at key to file.

        If a cache miss occurs, the value is retrieved, stored in the cache and
        returned.

        If the cache raises an :exc:`~IOError`, the retrieval cannot proceed: If
        ``file`` was an open file, data maybe been written to it already.
        The :exc:`~IOError` bubbles up.

        It is possible for a caching error to occur while attempting to store the value
        in the cache. It will not be handled as well.

        Parameters
        ----------
        key : str
            The key to be read.
        file : BinaryIO or str
            Output filename or file-like object with a ``write`` method.

        """
        try:
            return self.cache.get_file(key, file)
        except KeyError:
            # cache miss, load into cache
            fp = self._dstore.open(key)
            self.cache.put_file(key, fp)

            # return from cache
            return self.cache.get_file(key, file)
        # if an IOError occured, file pointer may be dirty - cannot proceed
        # safely

    def open(self, key: str) -> BinaryIO:
        """Open record at key.

        If a cache miss occurs, the value is retrieved, stored in the cache,
        then then another open is issued on the cache.

        If the cache raises an :exc:`~IOError`, the cache is
        ignored, and the backing store is consulted directly.

        It is possible for a caching error to occur while attempting to store
        the value in the cache. It will not be handled as well.

        Parameters
        ----------
        key : str
            Key to open.

        Returns
        -------
        file: BinaryIO
            Read-only file-like object for reading data at key.

        """
        try:
            return self.cache.open(key)
        except KeyError:
            # cache miss, load into cache
            fp = self._dstore.open(key)
            self.cache.put_file(key, fp)

            return self.cache.open(key)
        except OSError:
            # cache error, ignore completely and return from backend
            return self._dstore.open(key)

    def copy(self, source: str, dest: str) -> str:
        """Copy data at key ``source`` to key ``dest``.

        Copies the data in the backing store and removes the destination key from the
        cache, in case it was already populated.

        Parameters
        ----------
        source : str
            The source key of data to copy.
        dest : str
            The destination for the copy.

        Returns
        -------
        key : str
            The destination key.

        Raises
        ------
        ValueError
            If the underlying store does not support copy.
        """
        if not hasattr(self._dsctore, "copy"):
            raise ValueError(f"Store {type(self._dsctore)} does not support copy.")
        else:
            try:
                k = self._dstore.copy(source, dest)  # type: ignore
            finally:
                self.cache.delete(dest)
            return k

    def put(self, key: str, data: bytes) -> str:
        """Store bytestring data at key.

        Will store the value in the backing store. Afterwards delete the (original)
        value at key from the cache.

        Parameters
        ----------
        key : str
            The key under which the data is to be stored.
        data : bytes
            Data to be stored at key, must be of type  ``bytes``.

        Returns
        -------
        key: str
            The key under which data was stored.

        """
        try:
            return self._dstore.put(key, data)
        finally:
            self.cache.delete(key)

    def put_file(self, key: str, file: Union[str, BinaryIO]) -> str:
        """Store contents of file at key.

        Will store the value in the backing store. Afterwards delete the (original)
        value at key from the cache.

        Parameters
        ----------
        key : str
            Key where to store data in file.
        file : BinaryIO or str
            A filename or a file-like object with a read method.

        Returns
        -------
        key: str
            The key under which data was stored.

        """
        try:
            return self._dstore.put_file(key, file)
        finally:
            self.cache.delete(key)



import pytest

boto = pytest.importorskip("boto")
from bucket_manager import boto_bucket, boto_credentials


@pytest.fixture(
    params=boto_credentials, ids=[c["access_key"] for c in boto_credentials]
)
def credentials(request):
    return request.param


@pytest.fixture
def bucket(credentials):
    with boto_bucket(**credentials) as bucket:
        yield bucket


def test_simple(bucket):
    pass


def test_simple_with_contents(bucket):
    k = bucket.new_key("i_will_prevent_deletion")
    k.set_contents_from_string("meh")



import os
import os.path
from typing import TYPE_CHECKING, Any, Dict
from warnings import warn

from minimalkv.fs import FilesystemStore

if TYPE_CHECKING:
    from minimalkv._key_value_store import KeyValueStore


def create_store(type: str, params: Dict[str, Any]) -> "KeyValueStore":
    """Create store of type ``type`` with ``params``."""
    warn(
        """
        create_store will be removed in the next major release.
        If you want to create a KeyValueStore from a URL, use get_store_from_url.
        """,
        DeprecationWarning,
        stacklevel=2,
    )
    # TODO: More detailed docstring
    if type in ("azure", "hazure"):
        return _create_store_azure(type, params)
    if type in ("hs3", "boto"):
        return _create_store_hs3(type, params)
    if type in ("s3"):
        return _create_store_s3(type, params)
    if type in ("gcs", "hgcs"):
        return _create_store_gcs(type, params)
    if type in ("hfs", "hfile", "filesystem"):
        return _create_store_hfs(type, params)
    if type in ("fs", "file"):
        return _create_store_fs(type, params)
    if type in ("memory"):
        return _create_store_mem(type, params)
    if type in ("hmemory"):
        return _create_store_hmem(type, params)
    if type in ("redis"):
        return _create_store_redis(type, params)
    raise ValueError("Unknown store type: " + str(type))


def _create_store_gcs(store_type, params):
    # TODO: Docstring with required params.
    import json

    from google.oauth2.service_account import Credentials

    from minimalkv._hstores import HGoogleCloudStore
    from minimalkv.net.gcstore import GoogleCloudStore

    if isinstance(params["credentials"], bytes):
        account_info = json.loads(params["credentials"].decode())
        params["credentials"] = Credentials.from_service_account_info(
            account_info,
            scopes=["https://www.googleapis.com/auth/devstorage.read_write"],
        )
        params["project"] = account_info["project_id"]

    return (
        GoogleCloudStore(**params)
        if store_type == "gcs"
        else HGoogleCloudStore(**params)
    )


def _create_store_azure(type, params):
    # TODO: Docstring with required params.
    from minimalkv._hstores import HAzureBlockBlobStore
    from minimalkv.net.azurestore import AzureBlockBlobStore

    conn_string = params.get("connection_string", _build_azure_url(**params))

    if params["create_if_missing"] and params.get("use_sas", False):
        raise Exception("create_if_missing is incompatible with the use of SAS tokens.")

    if type == "azure":
        return AzureBlockBlobStore(
            conn_string=conn_string,
            container=params["container"],
            public=False,
            create_if_missing=params["create_if_missing"],
            checksum=params.get("checksum", True),
            max_connections=params.get("max_connections", 2),
            socket_timeout=params.get("socket_timeout", (20, 100)),
            max_block_size=params.get("max_block_size", (4194304)),
            max_single_put_size=params.get("max_single_put_size", (67108864)),
        )
    else:
        return HAzureBlockBlobStore(
            conn_string=conn_string,
            container=params["container"],
            public=False,
            create_if_missing=params["create_if_missing"],
            checksum=params.get("checksum", True),
            max_connections=params.get("max_connections", 2),
            socket_timeout=params.get("socket_timeout", (20, 100)),
            max_block_size=params.get("max_block_size", (4194304)),
            max_single_put_size=params.get("max_single_put_size", (67108864)),
        )


def _create_store_hs3(type, params):
    # TODO: Docstring with required params.
    from minimalkv._hstores import HBotoStore

    from ._boto import _get_s3bucket

    return HBotoStore(_get_s3bucket(**params))


def _create_store_s3(type, params):
    # TODO: Docstring with required params.
    from minimalkv.net.botostore import BotoStore

    from ._boto import _get_s3bucket

    return BotoStore(_get_s3bucket(**params))


def _create_store_hfs(type, params):
    # TODO: Docstring with required params.
    if params["create_if_missing"] and not os.path.exists(params["path"]):
        os.makedirs(params["path"])
    from minimalkv._hstores import HFilesystemStore

    return HFilesystemStore(params["path"])


def _create_store_fs(type, params):
    # TODO: Docstring with required params.
    if params["create_if_missing"] and not os.path.exists(params["path"]):
        os.makedirs(params["path"])
    return FilesystemStore(params["path"])


def _create_store_mem(type, params):
    # TODO: Docstring with required params.
    from minimalkv.memory import DictStore

    return DictStore()


def _create_store_hmem(type, params):
    # TODO: Docstring with required params.
    from minimalkv._hstores import HDictStore

    return HDictStore()


def _create_store_redis(type, params):
    # TODO: Docstring with required params.
    from redis import StrictRedis

    from minimalkv.memory.redisstore import RedisStore

    r = StrictRedis(**params)
    return RedisStore(r)


def _build_azure_url(
    account_name=None,
    account_key=None,
    default_endpoints_protocol=None,
    blob_endpoint=None,
    use_sas=False,
    **kwargs,
):
    # TODO: Docstring
    protocol = default_endpoints_protocol or "https"
    if use_sas:
        return (
            "DefaultEndpointsProtocol={protocol};AccountName={account_name};"
            "SharedAccessSignature={shared_access_signature}".format(
                protocol=protocol,
                account_name=account_name,
                shared_access_signature=account_key,
            )
        )
    else:
        return "DefaultEndpointsProtocol={protocol};AccountName={account_name};AccountKey={account_key}".format(
            protocol=protocol, account_name=account_name, account_key=account_key
        )



from minimalkv.decorator import ReadOnlyDecorator, URLEncodeKeysDecorator


def decorate_store(store, decoratorname):  # noqa D
    decoratorname_part = decoratorname.split("(")[0]
    if decoratorname_part == "urlencode":
        return URLEncodeKeysDecorator(store)
    if decoratorname_part == "readonly":
        return ReadOnlyDecorator(store)
    raise ValueError("Unknown store decorator: " + str(decoratorname))



"""Implement the AzureBlockBlobStore for `azure-storage-blob~=12`."""
import io
from contextlib import contextmanager
from typing import Optional

from minimalkv._key_value_store import KeyValueStore
from minimalkv.net._azurestore_common import _byte_buffer_md5, _file_md5
from minimalkv.net._net_common import LAZY_PROPERTY_ATTR_PREFIX, lazy_property


@contextmanager
def map_azure_exceptions(key=None, error_codes_pass=()):
    """Map Azure-specific exceptions to the minimalkv-API."""
    from azure.core.exceptions import AzureError

    try:
        yield
    except AzureError as ex:
        error_code = getattr(ex, "error_code", None)
        if error_code is not None and error_code in error_codes_pass:
            return
        if error_code == "BlobNotFound":
            raise KeyError(key) from ex
        raise OSError(str(ex)) from ex


class AzureBlockBlobStore(KeyValueStore):  # noqa D
    def __init__(
        self,
        conn_string=None,
        container=None,
        public=False,
        create_if_missing=True,
        max_connections=2,
        max_block_size=None,
        max_single_put_size=None,
        checksum=False,
        socket_timeout=None,
    ):
        from azure.storage.blob import BlobServiceClient, ContainerClient

        # Note that socket_timeout is unused; it only exist for backward compatibility.
        # TODO: Docstring
        self.conn_string = conn_string
        self.container = container
        self.public = public
        self.create_if_missing = create_if_missing
        self.max_connections = max_connections
        self.max_block_size = max_block_size
        self.max_single_put_size = max_single_put_size
        self.checksum = checksum
        self._service_client: Optional[BlobServiceClient] = None
        self._container_client: Optional[ContainerClient] = None

    # Using @lazy_property will (re-)create block_blob_service instance needed.
    # Together with the __getstate__ implementation below, this allows
    # AzureBlockBlobStore to be pickled, even if
    # azure.storage.blob.BlockBlobService does not support pickling.
    @lazy_property
    def blob_container_client(self):  # noqa D
        from azure.storage.blob import BlobServiceClient

        kwargs = {}
        if self.max_single_put_size:
            kwargs["max_single_put_size"] = self.max_single_put_size

        if self.max_block_size:
            kwargs["max_block_size"] = self.max_block_size

        self._service_client = BlobServiceClient.from_connection_string(
            self.conn_string, **kwargs
        )
        self._container_client = self._service_client.get_container_client(
            self.container
        )
        if self.create_if_missing:
            with map_azure_exceptions(error_codes_pass=("ContainerAlreadyExists")):
                self._container_client.create_container(
                    public_access="container" if self.public else None
                )
        return self._container_client

    def close(self):
        """Close container_client and service_client ports, if opened."""
        if self._container_client is not None:
            self._container_client.close()
            self._container_client = None
        if self._service_client is not None:
            self._service_client.close()
            self._service_client = None

    def _delete(self, key: str) -> None:
        with map_azure_exceptions(key, error_codes_pass=("BlobNotFound",)):
            self.blob_container_client.delete_blob(key)

    def _get(self, key):
        with map_azure_exceptions(key):
            blob_client = self.blob_container_client.get_blob_client(key)
            downloader = blob_client.download_blob(max_concurrency=self.max_connections)
            return downloader.readall()

    def _has_key(self, key):
        blob_client = self.blob_container_client.get_blob_client(key)
        with map_azure_exceptions(key, ("BlobNotFound",)):
            blob_client.get_blob_properties()
            return True
        return False

    def iter_keys(self, prefix=None):  # noqa D
        with map_azure_exceptions():
            blobs = self.blob_container_client.list_blobs(name_starts_with=prefix)

        def gen_names():  # noqa D
            with map_azure_exceptions():
                for blob in blobs:
                    yield blob.name

        return gen_names()

    def iter_prefixes(self, delimiter, prefix=""):  # noqa D
        return (
            blob_prefix.name
            for blob_prefix in self.blob_container_client.walk_blobs(
                name_starts_with=prefix, delimiter=delimiter
            )
        )

    def _open(self, key):
        with map_azure_exceptions(key):
            blob_client = self.blob_container_client.get_blob_client(key)
            return IOInterface(blob_client, self.max_connections)

    def _put(self, key, data):
        from azure.storage.blob import ContentSettings

        if self.checksum:
            content_settings = ContentSettings(
                content_md5=_byte_buffer_md5(data, b64encode=False)
            )
        else:
            content_settings = ContentSettings()

        with map_azure_exceptions(key):
            blob_client = self.blob_container_client.get_blob_client(key)

            blob_client.upload_blob(
                data,
                overwrite=True,
                content_settings=content_settings,
                max_concurrency=self.max_connections,
            )
        return key

    def _put_file(self, key, file):
        from azure.storage.blob import ContentSettings

        if self.checksum:
            content_settings = ContentSettings(
                content_md5=_file_md5(file, b64encode=False)
            )
        else:
            content_settings = ContentSettings()

        with map_azure_exceptions(key):
            blob_client = self.blob_container_client.get_blob_client(key)

            blob_client.upload_blob(
                file,
                overwrite=True,
                content_settings=content_settings,
                max_concurrency=self.max_connections,
            )
        return key

    def _get_file(self, key, file):
        with map_azure_exceptions(key):
            blob_client = self.blob_container_client.get_blob_client(key)
            downloader = blob_client.download_blob(max_concurrency=self.max_connections)
            downloader.readinto(file)

    def __getstate__(self):  # noqa D
        # keep all of __dict__, except lazy properties and properties, which need to be reopened:
        dont_pickle = {"_service_client", "_container_client"}
        return {
            key: value
            for key, value in self.__dict__.items()
            if not key.startswith(LAZY_PROPERTY_ATTR_PREFIX) and key not in dont_pickle
        }


class IOInterface(io.BufferedIOBase):
    """Class which provides a file-like interface to selectively read from a blob in the blob store."""

    def __init__(self, blob_client, max_connections):
        super().__init__()
        self.blob_client = blob_client
        self.max_connections = max_connections

        blob_props = self.blob_client.get_blob_properties()
        self.size = blob_props.size
        self.pos = 0

    def tell(self):
        """Return the current offset as int. Always >= 0."""
        if self.closed:
            raise ValueError("I/O operation on closed file")
        return self.pos

    def read(self, size=-1):
        """Return 'size' amount of bytes or less if there is no more data.

        If no size is given all data is returned. size can be >= 0.

        """
        if self.closed:
            raise ValueError("I/O operation on closed file")
        max_size = max(0, self.size - self.pos)
        if size < 0 or size > max_size:
            size = max_size
        if size == 0:
            return b""
        downloader = self.blob_client.download_blob(
            self.pos, size, max_concurrency=self.max_connections
        )
        b = downloader.readall()
        self.pos += len(b)
        return b

    def seek(self, offset, whence=0):
        """Move to a new offset either relative or absolute.

        whence=0 is absolute, whence=1 is relative, whence=2 is relative to the
        end.

        Any relative or absolute seek operation which would result in a
        negative position is undefined and that case can be ignored
        in the implementation.

        Any seek operation which moves the position after the stream
        should succeed. tell() should report that position and read()
        should return an empty bytes object.

        """
        if self.closed:
            raise ValueError("I/O operation on closed file")
        if whence == 0:
            if offset < 0:
                raise OSError("seek would move position outside the file")
            self.pos = offset
        elif whence == 1:
            if self.pos + offset < 0:
                raise OSError("seek would move position outside the file")
            self.pos += offset
        elif whence == 2:
            if self.size + offset < 0:
                raise OSError("seek would move position outside the file")
            self.pos = self.size + offset
        return self.pos

    def seekable(self):  # noqa D
        return True

    def readable(self):  # noqa D
        return True



from io import BytesIO

import pytest
from basic_store import BasicStore

from minimalkv.decorator import PrefixDecorator
from minimalkv.memory import DictStore


class TestPrefixDecorator(BasicStore):
    @pytest.fixture(
        params=[
            "short_",
            "loooooooooooooooooooooooooooooooooooooooooooooooooooooooooong_",
            "nounderscore",
            "_129073ashd812g",
        ]
    )
    def prefix(self, request):
        return request.param

    @pytest.fixture(
        params=[
            "prefix2_",
            "zz",
        ]
    )
    def prefix2(self, request):
        # these are used when multiple prefixes in a single store
        # are requested
        return request.param

    @pytest.fixture(params=[True, False])
    def store(self, request, prefix):
        base_store = DictStore()

        # do we add extra keys to the underlying store?
        if request.param:
            base_store.put("some_other_value", b"data1")
            base_store.put("ends_with_short_", b"data2")
            base_store.put("xx", b"data3")
            base_store.put("test", b"data4")

        return PrefixDecorator(prefix, base_store)

    def test_put_returns_correct_key(self, store, prefix, key, value):
        assert key == store.put(key, value)

    def test_put_sets_prefix(self, store, prefix, key, value):
        full_key = prefix + key
        assert key == store.put(key, value)

        assert store._dstore.get(full_key) == value

    def test_put_file_returns_correct_key(self, store, prefix, key, value):
        assert key == store.put_file(key, BytesIO(value))

    def test_put_file_sets_prefix(self, store, prefix, key, value):
        full_key = prefix + key
        assert key == store.put_file(key, BytesIO(value))

        assert store._dstore.get(full_key) == value

    def test_multiple_prefixes_one_store(self, store, prefix, prefix2, key, value):
        base_store = store._dstore
        store2 = PrefixDecorator(prefix2, base_store)

        pv = value + prefix.encode("ascii")
        pv2 = value + prefix2.encode("ascii")

        # put in with each prefix
        store.put(key, pv)
        store2.put(key, pv2)

        assert key in store
        assert key in store2

        assert prefix + key in base_store
        assert prefix2 + key in base_store

        assert len(store.keys()) == 1
        assert len(store2.keys()) == 1

        assert store.get(key) == pv
        assert store2.get(key) == pv2

    def test_pickle(self, store):
        import pickle

        rountrip = pickle.loads(pickle.dumps(store))
        assert isinstance(rountrip, PrefixDecorator)



from typing import Optional

from uritools import SplitResult


def _get_username(split_result: SplitResult) -> Optional[str]:
    userinfo = split_result.getuserinfo()
    if not userinfo:
        return None
    return userinfo.split(":")[0]


def _get_password(split_result: SplitResult) -> Optional[str]:
    userinfo = split_result.getuserinfo()
    if not userinfo or ":" not in userinfo:
        return None
    return userinfo.split(":")[1]



from io import BytesIO
from typing import BinaryIO, Iterator

from sqlalchemy import Column, LargeBinary, String, Table, exists, select
from sqlalchemy.orm import Session

from minimalkv import CopyMixin, KeyValueStore


class SQLAlchemyStore(KeyValueStore, CopyMixin):  # noqa D
    def __init__(self, bind, metadata, tablename):
        self.bind = bind

        self.table = Table(
            tablename,
            metadata,
            # 250 characters is the maximum key length that we guarantee can be
            # handled by any kind of backend
            Column("key", String(250), primary_key=True),
            Column("value", LargeBinary, nullable=False),
        )

    def _has_key(self, key: str) -> bool:
        with Session(self.bind) as session:
            return session.query(exists().where(self.table.c.key == key)).scalar()

    def _delete(self, key: str) -> None:
        with Session(self.bind) as session:
            session.execute(self.table.delete().where(self.table.c.key == key))
            session.commit()

    def _get(self, key: str) -> bytes:
        with Session(self.bind) as session:
            stmt = select(self.table.c.value).where(self.table.c.key == key)
            rv = session.execute(stmt).scalar()

            if not rv:
                raise KeyError(key)

            return rv

    def _open(self, key: str) -> BinaryIO:
        return BytesIO(self._get(key))

    def _copy(self, source: str, dest: str):
        with Session(self.bind) as session:
            # Find data at source key
            data_to_copy = self._get(source)

            # delete the potential existing previous key
            self.delete(dest)

            # insert new
            session.execute(
                self.table.insert().values(
                    {
                        "key": dest,
                        "value": data_to_copy,
                    }
                )
            )
            session.commit()
        return dest

    def _put(self, key: str, data: bytes) -> str:
        with Session(self.bind) as session:
            # delete the old
            session.execute(self.table.delete().where(self.table.c.key == key))

            # insert new
            session.execute(
                self.table.insert().values(
                    {
                        "key": key,
                        "value": data,
                    }
                )
            )
            session.commit()
        return key

    def _put_file(self, key: str, file: BinaryIO) -> str:
        return self._put(key, file.read())

    def iter_keys(self, prefix: str = "") -> Iterator[str]:  # noqa D
        with Session(self.bind) as session:
            query = select(self.table.c.key)
            if prefix != "":
                query = query.where(self.table.c.key.like(prefix + "%"))
            return (str(v[0]) for v in session.execute(query))



from functools import reduce
from typing import Any, Dict, List, Optional, Type

from uritools import SplitResult, urisplit

from minimalkv._key_value_store import KeyValueStore
from minimalkv._urls import url2dict


def get_store_from_url(
    url: str, store_cls: Optional[Type[KeyValueStore]] = None
) -> KeyValueStore:
    """Take a URL and return a minimalkv store according to the parameters in the URL.

    Parameters
    ----------
    url : str
        Access-URL, see below for supported formats.
    store_cls : Optional[Type[KeyValueStore]]
        The class of the store to create.
        If the URL scheme doesn't match the class, a ValueError is raised.

    Returns
    -------
    store : KeyValueStore
        KeyValueStore as described in url.

    Notes
    -----
    User credentials like secret keys have to be percent-encoded before they can be used
    in a URL (see ``azure`` and ``s3`` store types), since they can contain characters
    that are not valid in this part of a URL, like forward-slashes.

    You can use Python to percent-encode your secret key on the commandline like so::

        $ python -c "import urllib; print urllib.quote_plus('''dead/beef''')"
        dead%2Fbeef

    Store types and URL forms:

        * DictStore: ``memory://``
        * RedisStore: ``redis://[[password@]host[:port]][/db]``
        * FilesystemStore: ``fs://path``
        * BotoStore ``s3://access_key:secret_key@endpoint/bucket[?create_if_missing=true]``
        * AzureBlockBlockStorage: ``azure://account_name:account_key@container[?create_if_missing=true]``
        * AzureBlockBlockStorage (SAS): ``azure://account_name:shared_access_signature@container?use_sas&create_if_missing=false``
        * AzureBlockBlockStorage (SAS): ``azure://account_name:shared_access_signature@container?use_sas&create_if_missing=false[?max_connections=2&socket_timeout=(20,100)]``
        * AzureBlockBlockStorage (SAS): ``azure://account_name:shared_access_signature@container?use_sas&create_if_missing=false[?max_connections=2&socket_timeout=(20,100)][?max_block_size=4*1024*1024&max_single_put_size=64*1024*1024]``
        * GoogleCloudStorage: ``gcs://<base64 encoded credentials JSON>@bucket_name[?create_if_missing=true][&bucket_creation_location=EUROPE-WEST1]``
        * S3FSStore ``s3://access_key:secret_key@endpoint/bucket[?create_if_missing=true][&region_name=...]``

    See the respective store's :func:`_from_parsed_url` function for more details.

    """
    from minimalkv._hstores import HS3FSStore
    from minimalkv.net.s3fsstore import S3FSStore

    scheme_to_store: Dict[str, Type[KeyValueStore]] = {
        "s3": S3FSStore,
        "hs3": HS3FSStore,
        "boto": HS3FSStore,
    }

    parsed_url = urisplit(url)
    # Wrappers can be used to add functionality to a store, e.g. encryption.
    # See the documentation of _extract_wrappers for details.
    wrappers = _extract_wrappers(parsed_url)

    # Remove wrappers from scheme
    scheme_parts = parsed_url.getscheme().split("+")
    # pop off the type of the store
    scheme = scheme_parts[0]

    if scheme not in scheme_to_store:
        # If we can't find the scheme, we fall back to the old creation methods
        return get_store(**url2dict(url))

    store_cls_from_url = scheme_to_store[scheme]
    if store_cls is not None and store_cls_from_url != store_cls:
        raise ValueError(
            f"URL scheme {scheme} does not match store class {store_cls.__name__}"
        )

    query_listdict: Dict[str, List[str]] = parsed_url.getquerydict()
    # We will just use the last occurrence for each key
    query = {k: v[-1] for k, v in query_listdict.items()}

    store = store_cls_from_url._from_parsed_url(parsed_url, query)

    # apply wrappers/decorators:
    from minimalkv._store_decoration import decorate_store

    wrapped_store = reduce(decorate_store, wrappers, store)

    return wrapped_store


def _extract_wrappers(parsed_url: SplitResult) -> List[str]:
    """Extract wrappers from a parsed URL.

    Wrappers allow you to add additional functionality to a store, e.g. encryption.
    They can be specified in two ways:
    1. As the fragment part of the URL, e.g. "s3://...#wrap:readonly+urlencode"
    2. As part of the scheme, e.g. "s3+readonly+urlencode://..."

    The two methods cannot be mixed in the same URL.

    Parameters
    ----------
    parsed_url: SplitResult
        The parsed URL.

    Returns
    -------
    wrappers: List[str]
        The list of wrappers.
    """
    # Find wrappers in scheme, looking like this: "s3+readonly+urlencode://..."
    parts = parsed_url.getscheme().split("+")
    # pop off the type of the store
    parts.pop(0)
    scheme_wrappers = list(reversed(parts))

    # Find fragment wrappers, looking like this: "s3://...#wrap:readonly+urlencode"
    fragment = parsed_url.getfragment()
    fragments = fragment.split("#") if fragment else []
    wrap_spec = [s for s in fragments if s.startswith("wrap:")]
    if wrap_spec:
        fragment_without_wrap = wrap_spec[-1].partition("wrap:")[
            2
        ]  # remove the 'wrap:' part
        fragment_wrappers = list(fragment_without_wrap.split("+"))
    else:
        fragment_wrappers = []

    # can't have both:
    if scheme_wrappers and fragment_wrappers:
        raise ValueError(
            "Adding store wrappers via both the scheme and the fragment is not allowed."
        )

    return scheme_wrappers + fragment_wrappers


def get_store(
    type: str, create_if_missing: bool = True, **params: Any
) -> KeyValueStore:
    """Return a storage object according to the ``type`` and additional parameters.

    The ``type`` must be one of the types below, where each allows requires different
    parameters:

    * ``"azure"``: Returns a ``minimalkv.azure.AzureBlockBlobStorage``. Parameters are
      ``"account_name"``, ``"account_key"``, ``"container"``, ``"use_sas"`` and ``"create_if_missing"`` (default: ``True``).
      ``"create_if_missing"`` has to be ``False`` if ``"use_sas"`` is set. When ``"use_sas"`` is set,
      ``"account_key"`` is interpreted as Shared Access Signature (SAS) token.FIRE
      ``"max_connections"``: Maximum number of network connections used by one store (default: ``2``).
      ``"socket_timeout"``: maximum timeout value in seconds (socket_timeout: ``200``).
      ``"max_single_put_size"``: max_single_put_size is the largest size upload supported in a single put call.
      ``"max_block_size"``: maximum block size is maximum size of the blocks(maximum size is <= 100MB)
    * ``"s3"``: Returns a plain ``minimalkv.net.botostore.BotoStore``.
      Parameters must include ``"host"``, ``"bucket"``, ``"access_key"``, ``"secret_key"``.
      Optional parameters are

       - ``"force_bucket_suffix"`` (default: ``True``). If set, it is ensured that
         the bucket name ends with ``-<access_key>``
         by appending this string if necessary;
         If ``False``, the bucket name is used as-is.
       - ``"create_if_missing"`` (default: ``True`` ). If set, creates the bucket if it does not exist;
         otherwise, try to retrieve the bucket and fail with an ``IOError``.
    * ``"hs3"`` returns a variant of ``minimalkv.net.botostore.BotoStore`` that allows "/" in the key name.
      The parameters are the same as for ``"s3"``
    * ``"gcs"``: Returns a ``minimalkv.net.gcstore.GoogleCloudStore``.  Parameters are
      ``"credentials"``, ``"bucket_name"``, ``"bucket_creation_location"``, ``"project"`` and ``"create_if_missing"`` (default: ``True``).

      - ``"credentials"``: either the path to a credentials.json file or a *google.auth.credentials.Credentials* object
      - ``"bucket_name"``: Name of the bucket the blobs are stored in.
      - ``"project"``: The name of the GCStorage project. If a credentials JSON is passed then it contains the project name
        and this parameter will be ignored.
      - ``"create_if_missing"``: [optional] Create new bucket to store blobs in if ``"bucket_name"`` doesn't exist yet. (default: ``True``).
      - ``"bucket_creation_location"``: [optional] If a new bucket is created (create_if_missing=True), the location it will be created in.
        If ``None`` then GCloud uses a default location.
    * ``"hgcs"``: Like ``"gcs"`` but "/" are allowed in the keynames.
    * ``"fs"``: Returns a ``minimalkv.fs.FilesystemStore``. Specify the base path as "path" parameter.
    * ``"hfs"`` returns a variant of ``minimalkv.fs.FilesystemStore``  that allows "/" in the key name.
      The parameters are the same as for ``"file"``.
    * ``"memory"``: Returns a DictStore. Doesn't take any parameters
    * ``"redis"``: Returns a RedisStore. Constructs a StrictRedis using params as kwargs.
      See StrictRedis documentation for details.

    Parameters
    ----------
    type : str
        Type of storage to open, with optional storage decorators.
    create_if_missing : bool, optional, default = True
        Create the "root" of the storage (Azure container, parent directory, S3 bucket,
        etc.). Has no effect for stores where this makes no sense, like ``redis`` or
        ``memory``.
    kwargs
        Parameters specific to the store type.

    Returns
    -------
    store: KeyValueStore
        Key value store of type ``type`` as described in ``kwargs`` parameters.

    """
    from minimalkv._store_creation import create_store
    from minimalkv._store_decoration import decorate_store

    # split off old-style wrappers, if any:
    parts = type.split("+")
    type = parts.pop(0)
    decorators = list(reversed(parts))

    # find new-style wrappers, if any:
    wrapspec = params.pop("wrap", "")
    wrappers = list(wrapspec.split("+")) if wrapspec else []

    # can't have both:
    if wrappers:
        if decorators:
            raise ValueError(
                "Adding store wrappers via store type as well as via wrap parameter are not allowed. Preferably use wrap."
            )
        decorators = wrappers

    # create_if_missing is a universal parameter, so it's part of the function signature
    # it can be safely ignored by stores where 'creating' makes no sense.
    params["create_if_missing"] = create_if_missing

    store = create_store(type, params)

    # apply wrappers/decorators:
    wrapped_store = reduce(decorate_store, decorators, store)

    return wrapped_store



import hmac
import os
import tempfile
from io import BytesIO

import pytest

from minimalkv.crypt import HMACDecorator, VerificationException, _HMACFileReader


class TestHMACFileReader:
    @pytest.fixture
    def bad_datas(self, value):
        val = value * 3

        def _alter_byte(byte_string, pos):
            old_val = byte_string[pos]
            new_byte = bytes((old_val + 1 % 255,))
            return byte_string[:pos] + new_byte + byte_string[pos + 1 :]

        return (_alter_byte(val, i) for i in range(len(val)))

    @pytest.fixture
    def expected_digest(self, secret_key, value, hashfunc):
        return hmac.HMAC(secret_key, value, hashfunc).digest()

    @pytest.fixture
    def stored_blob(self, value, expected_digest):
        return value + expected_digest

    @pytest.fixture
    def create_reader(self, stored_blob, secret_key, hashfunc):
        return lambda: _HMACFileReader(
            hmac.HMAC(secret_key, None, hashfunc), BytesIO(stored_blob)
        )

    @pytest.fixture
    def chunk_sizes(self, value):
        return [10**n for n in range(2, 8)]

    def test_close(self, create_reader):
        reader = create_reader()
        assert not reader.source.closed
        reader.close()
        assert reader.source.closed

    def test_close_via_context(self, create_reader):
        reader = create_reader()
        assert not reader.source.closed
        with reader as r:
            assert r is reader
        assert reader.source.closed

    def test_reading_limit_0(self, create_reader):
        reader = create_reader()
        data = reader.read(0)
        assert isinstance(data, bytes)
        assert len(data) == 0
        data = reader.read(0)
        assert isinstance(data, bytes)
        assert len(data) == 0

    def test_reading_with_limit(
        self, secret_key, hashfunc, value, create_reader, chunk_sizes
    ):
        # try for different read lengths
        for n in chunk_sizes:
            chunks = []
            reader = create_reader()
            while True:
                r = reader.read(n)
                if not r:
                    break
                chunks.append(r)

            assert b"".join(chunks) == value

    def test_manipulated_input_full_read(self, secret_key, value, bad_datas, hashfunc):
        for bad_data in bad_datas:
            reader = _HMACFileReader(
                hmac.HMAC(secret_key, None, hashfunc), BytesIO(bad_data)
            )

            with pytest.raises(VerificationException):
                reader.read()

    def test_manipulated_input_incremental_read(self, secret_key, bad_datas, hashfunc):
        for bad_data in bad_datas:
            reader = _HMACFileReader(
                hmac.HMAC(secret_key, None, hashfunc), BytesIO(bad_data)
            )

            with pytest.raises(VerificationException):
                bitesize = 100
                while True:
                    if len(reader.read(bitesize)) != bitesize:
                        break

    def test_input_too_short(self, secret_key, hashfunc):
        with pytest.raises(VerificationException):
            _HMACFileReader(hmac.HMAC(secret_key, None, hashfunc), BytesIO(b"a"))

    def test_unbounded_read(self, value, create_reader):
        assert create_reader().read() == value


# test the "real" HMACMixin: core functionality and checks
# this only works with dicts, as we access the internal structures to
# manipulate values
class HMACDec:
    @pytest.fixture
    def hmacstore(self, secret_key, store):
        return HMACDecorator(secret_key, store)

    def test_get_fails_on_manipulation(self, hmacstore, key, value):
        hmacstore.put(key, value)
        hmacstore.d[key] += b"a"

        with pytest.raises(VerificationException):
            hmacstore.get(key)

    def test_copy_raises_not_implemented(self, store):
        with pytest.raises(NotImplementedError):
            HMACDecorator(b"secret", store).copy("src", "dest")

    def test_put_file_obj(self, key, value, hmacstore):
        hmacstore.put_file(key, BytesIO(value))
        assert hmacstore.get(key) == value

    def test_put_file_str(self, key, value, hmacstore):
        with tempfile.NamedTemporaryFile(mode="wb", delete=False) as f:
            f.write(value)
        hmacstore.put_file(key, f.name)
        assert hmacstore.get(key) == value

    def test_get_file_obj(self, key, value, hmacstore):
        hmacstore.put(key, value)
        b = BytesIO()
        hmacstore.get_file(key, b)
        assert b.getvalue() == value

    def test_get_file_non_writable_target(self, key, value, hmacstore):
        hmacstore.put(key, value)
        path = "/tmp/this/file/does/not/exist"
        with pytest.raises(IOError, match=f"Error opening {path} for writing"):
            hmacstore.get_file(key, path)

    def test_get_file_fails_on_manipulation(self, hmacstore, key, value):
        hmacstore.put(key, value)
        hmacstore.d[key] += b"a"

        with tempfile.TemporaryFile() as tmp:
            with pytest.raises(VerificationException):
                hmacstore.get_file(key, tmp)

        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            try:
                with pytest.raises(VerificationException):
                    hmacstore.get_file(key, tmp.name)
            finally:
                os.unlink(tmp.name)

    def test_open_fails_on_manipulation(self, hmacstore, key, value):
        hmacstore.put(key, value)
        hmacstore.d[key] += b"a"

        with pytest.raises(VerificationException):
            hmacstore.open(key).read()

        handle = hmacstore.open(key)

        # we read 1 extra byte now, because the value is actually longer
        handle.read(len(value) + 1)

        with pytest.raises(VerificationException):
            handle.read(1)

    def test_get_fails_on_replay_manipulation(self, hmacstore, key, key2, value):
        hmacstore.put(key, value)
        hmacstore.d[key2] = hmacstore.d[key]
        hmacstore.get(key)

        with pytest.raises(VerificationException):
            hmacstore.get(key2)



from typing import Callable

import pytest

from minimalkv._get_store import (
    get_store,
    get_store_from_url as get_store_from_url_new,
)
from minimalkv._hstores import HDictStore
from minimalkv._key_value_store import KeyValueStore
from minimalkv._urls import url2dict
from minimalkv.memory import DictStore


# Monkey patch equality operator for testing purposes.
def _eq_dict_store(self: object, other: object) -> bool:
    if isinstance(self, DictStore):
        if isinstance(other, DictStore):
            return self.d == other.d
    raise NotImplementedError


DictStore.__eq__ = _eq_dict_store  # type: ignore


def get_store_from_url_old(url: str) -> KeyValueStore:
    return get_store(**url2dict(url))


@pytest.fixture(params=[get_store_from_url_new, get_store_from_url_old])
def get_store_from_url(request) -> Callable:
    return request.param


@pytest.mark.parametrize(
    "url, key_value_store", [("memory://", DictStore()), ("hmemory://", HDictStore())]
)
def test_get_store_from_url(
    url: str, key_value_store: KeyValueStore, get_store_from_url: Callable
) -> None:
    assert get_store_from_url(url) == key_value_store



#!/usr/bin/env python
import re
from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO, Iterator, List, Optional, Union

if TYPE_CHECKING:
    from redis import StrictRedis

from minimalkv._constants import FOREVER, NOT_SET
from minimalkv._key_value_store import KeyValueStore
from minimalkv._mixins import TimeToLiveMixin


class RedisStore(TimeToLiveMixin, KeyValueStore):
    """Uses a redis-database as the backend.

    Parameters
    ----------
    redis : redis.StrictRedis
        Backend.

    """

    def __init__(self, redis: "StrictRedis"):
        self.redis = redis

    def _delete(self, key: str) -> int:
        return self.redis.delete(key)

    def keys(self, prefix: str = "") -> List[str]:
        """List all keys in the store starting with prefix.

        Parameters
        ----------
        prefix : str, optional, default = ''
            Only list keys starting with prefix. List all keys if empty.

        Raises
        ------
        IOError
            If there was an error accessing the store.
        """
        return [b.decode() for b in self.redis.keys(pattern=re.escape(prefix) + "*")]

    def iter_keys(self, prefix="") -> Iterator[str]:
        """Iterate over all keys in the store starting with prefix.

        Parameters
        ----------
        prefix : str, optional, default = ''
            Only iterate over keys starting with prefix. Iterate over all keys if empty.

        """
        return iter(self.keys(prefix))

    def _has_key(self, key: str) -> bool:
        return self.redis.exists(key) > 0

    def _get(self, key: str) -> bytes:
        val = self.redis.get(key)

        if val is None:
            raise KeyError(key)
        return val

    def _get_file(self, key: str, file: BinaryIO) -> str:
        file.write(self._get(key))
        return key

    def _open(self, key: str) -> BinaryIO:
        return BytesIO(self._get(key))

    def _put(
        self, key: str, value: bytes, ttl_secs: Optional[Union[str, int, float]] = None
    ) -> str:
        assert ttl_secs is not None
        if ttl_secs in (NOT_SET, FOREVER):
            # if we do not care about ttl, just use set
            # in redis, using SET will also clear the timeout
            # note that this assumes that there is no way in redis
            # to set a default timeout on keys
            self.redis.set(key, value)
        else:
            ittl = None
            try:
                ittl = int(ttl_secs)
            except ValueError:
                pass  # let it blow up further down

            if ittl == ttl_secs:
                self.redis.setex(key, ittl, value)
            else:
                self.redis.psetex(key, int(ttl_secs * 1000), value)

        return key

    def _put_file(
        self,
        key: str,
        file: BinaryIO,
        ttl_secs: Optional[Union[str, int, float]] = None,
    ) -> str:
        self._put(key, file.read(), ttl_secs)
        return key



from minimalkv._constants import (
    FOREVER,
    NOT_SET,
    VALID_KEY_RE,
    VALID_KEY_REGEXP,
    VALID_NON_NUM,
)
from minimalkv._get_store import get_store, get_store_from_url
from minimalkv._key_value_store import KeyValueStore, UrlKeyValueStore
from minimalkv._mixins import CopyMixin, TimeToLiveMixin, UrlMixin
from minimalkv._store_creation import create_store
from minimalkv._store_decoration import decorate_store
from minimalkv._urls import url2dict

try:
    import importlib.metadata

    __version__ = importlib.metadata.version(__name__)
except Exception:
    __version__ = "unknown"


__all__ = [
    "CopyMixin",
    "create_store",
    "decorate_store",
    "FOREVER",
    "get_store_from_url",
    "get_store",
    "KeyValueStore",
    "NOT_SET",
    "TimeToLiveMixin",
    "url2dict",
    "UrlKeyValueStore",
    "UrlMixin",
    "VALID_KEY_RE",
    "VALID_KEY_REGEXP",
    "VALID_NON_NUM",
]



from unittest import mock

from minimalkv import KeyValueStore


def test_keyvaluestore_enter_exit():
    with mock.patch("minimalkv.KeyValueStore.close") as closefunc:
        with KeyValueStore() as kv:  # noqa F841
            pass
        closefunc.assert_called_once()



from minimalkv._constants import (
    VALID_KEY_RE_EXTENDED,
    VALID_KEY_REGEXP_EXTENDED,
    VALID_NON_NUM_EXTENDED,
)
from minimalkv._mixins import ExtendedKeyspaceMixin

__all__ = [
    "VALID_NON_NUM_EXTENDED",
    "VALID_KEY_REGEXP_EXTENDED",
    "VALID_KEY_RE_EXTENDED",
    "ExtendedKeyspaceMixin",
]



import os
import warnings
from typing import Dict, NamedTuple, Optional

from uritools import SplitResult

from minimalkv import UrlMixin
from minimalkv._url_utils import _get_password, _get_username
from minimalkv.fsspecstore import FSSpecStore

try:
    from s3fs import S3FileSystem

    has_s3fs = True
except ImportError:
    has_s3fs = False

warnings.warn(
    "This class will be renamed to `Boto3Store` in the next major release.",
    category=DeprecationWarning,
    stacklevel=2,
)


class Credentials(NamedTuple):
    """Dataclass to hold AWS credentials."""

    access_key_id: Optional[str]
    secret_access_key: Optional[str]
    session_token: Optional[str]

    def as_boto3_params(self):
        """Return the credentials as a dictionary suitable for boto3 authentication."""
        return {
            "aws_access_key_id": self.access_key_id,
            "aws_secret_access_key": self.secret_access_key,
            "aws_session_token": self.session_token,
        }


class S3FSStore(FSSpecStore, UrlMixin):  # noqa D
    def __init__(
        self,
        bucket,
        credentials: Optional[Credentials] = None,
        object_prefix="",
        url_valid_time=0,
        reduced_redundancy=False,
        public=False,
        metadata=None,
        verify=True,
        region_name=None,
    ):
        if isinstance(bucket, str):
            import boto3

            boto3_params = credentials.as_boto3_params() if credentials else {}
            s3_resource = boto3.resource("s3", **boto3_params)
            bucket = s3_resource.Bucket(bucket)
            if bucket not in s3_resource.buckets.all():
                raise ValueError("invalid s3 bucket name")

        self.bucket = bucket
        self.credentials = credentials
        self.object_prefix = object_prefix.strip().lstrip("/")
        self.url_valid_time = url_valid_time
        self.reduced_redundancy = reduced_redundancy
        self.public = public
        self.metadata = metadata or {}
        self.verify = verify
        self.region_name = region_name

        # Get endpoint URL
        self.endpoint_url = self.bucket.meta.client.meta.endpoint_url

        write_kwargs = {
            "Metadata": self.metadata,
        }
        if self.reduced_redundancy:
            write_kwargs["StorageClass"] = "REDUCED_REDUNDANCY"
        if self.public:
            write_kwargs["ACL"] = "public-read"
        super().__init__(
            prefix=f"{bucket.name}/{self.object_prefix}", write_kwargs=write_kwargs
        )

    def _create_filesystem(self) -> "S3FileSystem":
        if not has_s3fs:
            raise ImportError("Cannot find optional dependency s3fs.")

        client_kwargs = {"verify": self.verify}
        if self.endpoint_url:
            client_kwargs["endpoint_url"] = self.endpoint_url
        if self.region_name:
            client_kwargs["region_name"] = self.region_name

        if self.credentials:
            return S3FileSystem(
                key=self.credentials.access_key_id,
                secret=self.credentials.secret_access_key,
                token=self.credentials.session_token,
                anon=False,
                client_kwargs=client_kwargs,
            )

        return S3FileSystem(
            anon=False,
            client_kwargs=client_kwargs,
        )

    def _url_for(self, key) -> str:
        return self._fs.url(
            f"{self.bucket.name}/{self.object_prefix}{key}", expires=self.url_valid_time
        )

    @classmethod
    def _from_parsed_url(
        cls, parsed_url: SplitResult, query: Dict[str, str]
    ) -> "S3FSStore":  # noqa D
        """Build an ``S3FSStore`` from a parsed URL.

        To build an ``S3FSStore`` from a URL, use :func:`get_store_from_url`.

        URl format:
        ``s3://access_key_id:secret_access_key@endpoint/bucket[?<query_args>]``

        **Positional arguments**:

        ``access_key_id``: The access key ID of the S3 user.

        ``secret_access_key``: The secret access key of the S3 user.

        ``endpoint``: The endpoint of the S3 service. Leave empty for standard AWS.

        ``bucket``: The name of the bucket.

        **Query arguments**:

        ``force_bucket_suffix`` (default: ``True``): If set, it is ensured that
         the bucket name ends with ``-<access_key>``
         by appending this string if necessary.
         If ``False``, the bucket name is used as-is.

        ``create_if_missing`` (default: ``True`` ): If set, creates the bucket if it does not exist;
         otherwise, try to retrieve the bucket and fail with an ``IOError``.

        ``region_name`` (default: ``None``): If set the AWS region name is applied as location
        constraint during bucket creation.

        ``session_token``(default: ``None``): If set this token will be used in conjunction
        with access_key_id and secret_access_key for authentication.

        **Notes**:

        If the scheme is ``hs3``, an ``HS3FSStore`` is returned which allows ``/`` in key names.

        If the credentials are not provided through the url, they are attempted to be
        loaded from the environment variables `AWS_ACCESS_KEY_ID`,
        `AWS_SECRET_ACCESS_KEY`, and `AWS_SESSION_TOKEN`. If these variables are not set,
        the search for credentials will be delegated to boto(core).

        Positional arguments should be encoded by `urllib.parse.quote_plus`
        if they contain special characters e.g. "/".

        Parameters
        ----------
        parsed_url: SplitResult
            The parsed URL.
        query: Dict[str, str]
            Query parameters from the URL.

        Returns
        -------
        store : S3FSStore
            The created S3FSStore.
        """
        import boto3

        url_access_key_id = _get_username(parsed_url)
        url_secret_access_key = _get_password(parsed_url)
        url_session_token = query.get("session_token", None)

        if url_access_key_id is None:
            url_secret_access_key = os.environ.get("AWS_ACCESS_KEY_ID")
        else:
            os.environ["AWS_ACCESS_KEY_ID"] = url_access_key_id

        if url_secret_access_key is None:
            url_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY")
        else:
            os.environ["AWS_SECRET_ACCESS_KEY"] = url_secret_access_key

        credentials = Credentials(
            access_key_id=url_access_key_id,
            secret_access_key=url_secret_access_key,
            session_token=url_session_token,
        )

        boto3_params = credentials.as_boto3_params()
        host = parsed_url.gethost()
        port = parsed_url.getport()

        is_secure = query.get("is_secure", "true").lower() == "true"
        endpoint_scheme = "https" if is_secure else "http"

        if host is None:
            endpoint_url = None
        elif port is None:
            endpoint_url = f"{endpoint_scheme}://{host}"
        else:
            endpoint_url = f"{endpoint_scheme}://{host}:{port}"

        boto3_params["endpoint_url"] = endpoint_url

        # Remove Nones from client_params
        boto3_params = {k: v for k, v in boto3_params.items() if v is not None}

        bucket_name = parsed_url.getpath().lstrip("/")

        resource = boto3.resource("s3", **boto3_params)  # type: ignore

        force_bucket_suffix = query.get("force_bucket_suffix", "true").lower() == "true"
        if force_bucket_suffix:
            # Try to find access key in env
            if url_access_key_id is None:
                access_key_id = os.environ.get("AWS_ACCESS_KEY_ID")
            else:
                access_key_id = url_access_key_id

            if access_key_id is None:
                raise ValueError(
                    "Cannot find access key in URL or environment variable AWS_ACCESS_KEY_ID"
                )

            if not bucket_name.lower().endswith("-" + access_key_id.lower()):
                bucket_name += "-" + access_key_id.lower()

        # We only create a reference to the bucket here.
        # The bucket will be created in the `create_filesystem` method if it doesn't exist.
        region_name = query.get("region_name")

        bucket = resource.Bucket(bucket_name)

        verify = query.get("verify", "true").lower() == "true"

        return cls(
            bucket, credentials=credentials, verify=verify, region_name=region_name
        )



import os

import pytest

boto = pytest.importorskip("boto")
from io import BytesIO

from basic_store import BasicStore
from bucket_manager import boto_bucket, boto_credentials
from conftest import ExtendedKeyspaceTests
from url_store import UrlStore

from minimalkv._mixins import ExtendedKeyspaceMixin
from minimalkv.net.botostore import BotoStore


@pytest.fixture(
    params=boto_credentials, ids=[c["access_key"] for c in boto_credentials]
)
def credentials(request):
    return request.param


@pytest.fixture
def bucket(credentials):
    with boto_bucket(**credentials) as bucket:
        yield bucket


class TestBotoStorage(BasicStore, UrlStore):
    @pytest.fixture(params=[True, False])
    def reduced_redundancy(self, request):
        return request.param

    @pytest.fixture
    def storage_class(self, reduced_redundancy):
        return "REDUCED_REDUNDANCY" if reduced_redundancy else "STANDARD"

    @pytest.fixture(params=["", "/test-prefix"])
    def prefix(self, request):
        return request.param

    @pytest.fixture
    def store(self, bucket, prefix, reduced_redundancy):
        return BotoStore(bucket, prefix, reduced_redundancy=reduced_redundancy)

    # Disable max key length test as it leads to problems with minio
    test_max_key_length = None

    def test_get_filename_nonexistant(self, store, key, tmp_path):
        # NOTE: boto misbehaves here and tries to erase the target file
        # the parent tests use /dev/null, which you really should not try
        # to os.remove!
        with pytest.raises(KeyError):
            store.get_file(key, os.path.join(str(tmp_path), "a"))

    def test_key_error_on_nonexistant_get_filename(self, store, key, tmp_path):
        with pytest.raises(KeyError):
            store.get_file(key, os.path.join(str(tmp_path), "a"))

    def test_storage_class_put(self, store, prefix, key, value, storage_class, bucket):
        store.put(key, value)

        keyname = prefix + key

        if storage_class != "STANDARD":
            pytest.xfail("boto does not support checking the storage class?")

        assert bucket.lookup(keyname).storage_class == storage_class

    def test_storage_class_putfile(
        self, store, prefix, key, value, storage_class, bucket
    ):
        store.put_file(key, BytesIO(value))

        keyname = prefix + key

        if storage_class != "STANDARD":
            pytest.xfail("boto does not support checking the storage class?")
        assert bucket.lookup(keyname).storage_class == storage_class


class TestExtendedKeyspaceBotoStore(TestBotoStorage, ExtendedKeyspaceTests):
    @pytest.fixture
    def store(self, bucket, prefix, reduced_redundancy):
        class ExtendedKeyspaceStore(ExtendedKeyspaceMixin, BotoStore):
            pass

        return ExtendedKeyspaceStore(
            bucket, prefix, reduced_redundancy=reduced_redundancy
        )



import os

from minimalkv._mixins import ExtendedKeyspaceMixin
from minimalkv.fs import FilesystemStore
from minimalkv.memory import DictStore
from minimalkv.memory.redisstore import RedisStore
from minimalkv.net.azurestore import AzureBlockBlobStore
from minimalkv.net.boto3store import Boto3Store
from minimalkv.net.botostore import BotoStore
from minimalkv.net.gcstore import GoogleCloudStore
from minimalkv.net.s3fsstore import S3FSStore


class HDictStore(ExtendedKeyspaceMixin, DictStore):  # noqa D
    pass


class HRedisStore(ExtendedKeyspaceMixin, RedisStore):  # noqa D
    pass


class HAzureBlockBlobStore(ExtendedKeyspaceMixin, AzureBlockBlobStore):  # noqa D
    pass


class HBotoStore(ExtendedKeyspaceMixin, BotoStore):  # noqa D
    def size(self, key: str) -> bytes:
        """Get size of data at key in bytes.

        Parameters
        ----------
        key : str
            Key of data.

        Returns
        -------
        size : int
            Size of value at key in bytes.
        """
        k = self.bucket.lookup(self.prefix + key)
        return k.size


class HS3FSStore(ExtendedKeyspaceMixin, S3FSStore):  # noqa D
    pass


class HBoto3Store(ExtendedKeyspaceMixin, Boto3Store):  # noqa D
    pass


class HGoogleCloudStore(ExtendedKeyspaceMixin, GoogleCloudStore):  # noqa D
    pass


class HFilesystemStore(ExtendedKeyspaceMixin, FilesystemStore):  # noqa D
    def size(self, key: str) -> int:
        """Get size of data at key in bytes.

        Parameters
        ----------
        key : str
            Key of data.

        Returns
        -------
        size : int
            Size of value at key in bytes.
        """
        return os.path.getsize(self._build_filename(key))



"""Internal utilities for ``aztorestore_old`` and ``azurestore_new``."""

import base64
import hashlib


def _file_md5(file_, b64encode=True):
    """Compute the md5 digest of a file in base64 encoding.

    For ``b64encode``, returns the base64 encoded string; otherwise, returns the
    bytes directly.
    """
    md5 = hashlib.md5()
    chunk_size = 128 * md5.block_size
    for chunk in iter(lambda: file_.read(chunk_size), b""):
        md5.update(chunk)
    file_.seek(0)
    byte_digest = md5.digest()
    if b64encode:
        return base64.b64encode(byte_digest).decode()
    else:
        return byte_digest


def _filename_md5(filename, b64encode=True):
    """Compute the md5 digest of a file in base64 encoding."""
    with open(filename, "rb") as f:
        return _file_md5(f, b64encode=b64encode)


def _byte_buffer_md5(buffer_, b64encode=True):
    """Compute the md5 digest of a byte buffer in base64 encoding."""
    md5 = hashlib.md5(buffer_)
    byte_digest = md5.digest()
    if b64encode:
        return base64.b64encode(byte_digest).decode()
    else:
        return byte_digest



import re
import time
from io import BytesIO
from typing import BinaryIO, Iterator, List, Optional, Union

from dulwich.objects import Blob, Commit, Tree
from dulwich.repo import Repo

from minimalkv import __version__
from minimalkv._key_value_store import KeyValueStore


def _on_tree(
    repo: Repo,
    tree: Tree,
    components: List[bytes],
    obj: Optional[Union[Blob, Tree]],
) -> List[Tree]:
    """Mount an object on a tree, using the given path components.

    Parameters
    ----------
    repo : dulwich.objects.Repo
        Repository.
    tree : dulwich.objects.Tree
        Tree object to mount on.
    components : list of str
        A list of strings (or bytes) of subpaths (e.g. ['foo', 'bar'] is equivalent to
        '/foo/bar').
    obj : dulwich.objects.Blob or dulwich.objects.Tree or None
        Object to mount. If None, removes the object found at path and prunes the tree
        downwards.

    Returns
    -------
    list of dulwich.objects:
        A list of new entities that need to be added to the object store, where the last
        one is the new tree.

    """
    # pattern-matching:
    if len(components) == 1:
        if isinstance(obj, Blob):
            mode: Optional[int] = 0o100644
        elif isinstance(obj, Tree):
            mode = 0o040000
        elif obj is None:
            mode = None
        else:
            raise TypeError("Can only mount Blobs or Trees")
        name = components[0]

        if mode is not None:
            assert obj is not None
            tree[name] = mode, obj.id
            return [tree]
        if name in tree:
            del tree[name]
        return [tree]
    elif len(components) > 1:
        a, bc = components[0], components[1:]
        if a in tree:
            a_tree = repo[tree[a][1]]
            if not isinstance(a_tree, Tree):
                a_tree = Tree()
        else:
            a_tree = Tree()
        res = _on_tree(repo, a_tree, bc, obj)
        a_tree_new = res[-1]

        if a_tree_new.items():
            tree[a] = 0o040000, a_tree_new.id
            return res + [tree]

        # tree is empty
        if a in tree:
            del tree[a]
        return [tree]
    else:
        raise ValueError("Components can't be empty.")


class GitCommitStore(KeyValueStore):
    """Store using git.

    Parameters
    ----------
    repo_path : str
        Path to the repository.
    branch : bytes, optional, default = b"master"
        Branch to use.
    subdir : bytes, optional, default = b""
        Subdirectory of the repository to use.

    """

    AUTHOR = f"GitCommitStore (minimalkv {__version__}) <>"
    TIMEZONE = None

    def __init__(self, repo_path: str, branch: bytes = b"master", subdir: bytes = b""):
        self.repo = Repo(repo_path)
        self.branch = branch

        # cleans up subdir, to a form of 'a/b/c' (no duplicate, leading or
        # trailing slashes)
        self.subdir = re.sub("#/+#", "/", subdir.decode("ascii").strip("/"))

    @property
    def _subdir_components(self) -> List[bytes]:
        return [c.encode("ascii") for c in self.subdir.split("/")]

    def _key_components(self, key: str) -> List[bytes]:
        return [c.encode("ascii") for c in key.split("/")]

    @property
    def _refname(self):
        return b"refs/heads/" + self.branch

    def _create_top_commit(self):
        # get the top commit, create empty one if it does not exist
        commit = Commit()

        # commit metadata
        author = self.AUTHOR.encode("utf8")
        commit.author = commit.committer = author
        commit.commit_time = commit.author_time = int(time.time())

        if self.TIMEZONE is not None:
            tz = self.TIMEZONE
        else:
            tz = time.timezone if (time.localtime().tm_isdst) else time.altzone
        commit.commit_timezone = commit.author_timezone = tz
        commit.encoding = b"UTF-8"

        return commit

    def _delete(self, key: str) -> None:
        try:
            commit = self.repo[self._refname]
            tree = self.repo[commit.tree]
        except KeyError:
            return  # not-found key errors are ignored

        commit = self._create_top_commit()
        objects_to_add = []

        components = self._key_components(key)
        if self.subdir:
            components = self._subdir_components + components

        res = _on_tree(self.repo, tree, components, None)
        objects_to_add.extend(res)
        tree = res[-1]

        commit.tree = tree.id
        commit.message = ("Deleted key {}".format(self.subdir + "/" + key)).encode(
            "utf8"
        )

        objects_to_add.append(commit)

        for obj in objects_to_add:
            self.repo.object_store.add_object(obj)

        self.repo.refs[self._refname] = commit.id

    def _get(self, key: str) -> bytes:
        # might raise key errors, except block corrects param
        try:
            commit = self.repo[self._refname]
            tree = self.repo[commit.tree]
            fn = self.subdir + "/" + key
            _, blob_id = tree.lookup_path(self.repo.__getitem__, fn.encode("ascii"))
            blob = self.repo[blob_id]
        except KeyError as e:
            raise KeyError(key) from e

        return blob.data

    def iter_keys(self, prefix: str = "") -> Iterator[str]:  # noqa D
        try:
            commit = self.repo[self._refname]
            tree = self.repo[commit.tree]

            if self.subdir:
                tree = self.repo[
                    tree.lookup_path(
                        self.repo.__getitem__, self.subdir.encode("ascii")
                    )[1]
                ]
        except KeyError:
            pass
        else:
            for o in self.repo.object_store.iter_tree_contents(
                tree.sha().hexdigest().encode("ascii")
            ):
                if o.path.decode("ascii").startswith(prefix):
                    yield o.path.decode("ascii")

    def _open(self, key: str) -> BinaryIO:
        return BytesIO(self._get(key))

    def _put_file(self, key: str, file: BinaryIO) -> str:
        # FIXME: it may be worth to try to move large files directly into the
        #        store here
        return self._put(key, file.read())

    def _put(self, key: str, data: bytes) -> str:
        commit = self._create_top_commit()
        commit.message = ("Updated key {}".format(self.subdir + "/" + key)).encode(
            "utf8"
        )

        blob = Blob.from_string(data)

        try:
            parent_commit = self.repo[self._refname]
        except KeyError:
            # branch does not exist, start with an empty tree
            tree = Tree()
        else:
            commit.parents = [parent_commit.id]
            tree = self.repo[parent_commit.tree]

        objects_to_add = [blob]

        components = self._key_components(key)
        if self.subdir:
            components = self._subdir_components + components
        res = _on_tree(self.repo, tree, components, blob)
        objects_to_add.extend(res)

        commit.tree = res[-1].id
        objects_to_add.append(commit)

        # add objects
        for obj in objects_to_add:
            self.repo.object_store.add_object(obj)

        # update refs
        self.repo.refs[self._refname] = commit.id

        return key



def _get_s3bucket(
    host,
    bucket,
    access_key,
    secret_key,
    force_bucket_suffix=True,
    create_if_missing=True,
):
    # TODO: Write docstring.
    from boto.s3.connection import (  # type: ignore
        OrdinaryCallingFormat,
        S3Connection,
        S3ResponseError,
    )

    s3_connection_params = {
        "aws_access_key_id": access_key,
        "aws_secret_access_key": secret_key,
        "is_secure": False,
        "calling_format": OrdinaryCallingFormat(),
    }

    # Split up the host into host and port.
    if ":" in host:
        host, port = host.split(":")
        s3_connection_params["port"] = int(port)
    s3_connection_params["host"] = host

    s3con = S3Connection(**s3_connection_params)

    # add access key prefix to bucket name, unless explicitly prohibited
    if force_bucket_suffix and not bucket.lower().endswith("-" + access_key.lower()):
        bucket = bucket + "-" + access_key.lower()
    try:
        return s3con.get_bucket(bucket)
    except S3ResponseError as ex:
        if ex.status == 404:
            if create_if_missing:
                return s3con.create_bucket(bucket)
            else:
                raise OSError(f"Bucket {bucket} does not exist") from ex
        raise



from io import BytesIO
from typing import Dict, Iterator, Optional

from uritools import SplitResult

from minimalkv import CopyMixin, KeyValueStore


class DictStore(KeyValueStore, CopyMixin):
    """Store data in a dictionary.

    This store uses a dictionary as the backend for storing, its implementation
    is straightforward. The dictionary containing all data is available as `d`.

    """

    d: Dict[str, bytes]

    def __init__(self, d: Optional[Dict[str, bytes]] = None):
        self.d = d or {}

    def _delete(self, key: str) -> None:
        self.d.pop(key, None)

    def _has_key(self, key: str) -> bool:
        return key in self.d

    def _open(self, key: str):
        return BytesIO(self.d[key])

    def _copy(self, source: str, dest: str) -> None:
        self.d[dest] = self.d[source]

    def _put_file(self, key: str, file, *args, **kwargs) -> str:
        self.d[key] = file.read()
        return key

    def iter_keys(self, prefix: str = "") -> Iterator[str]:
        """Iterate over all keys in the store starting with prefix.

        Parameters
        ----------
        prefix : str, optional, default = ''
            Only iterate over keys starting with prefix. Iterate over all keys if empty.

        """
        return filter(lambda k: k.startswith(prefix), iter(self.d))

    @classmethod
    def _from_parsed_url(
        cls, parsed_url: SplitResult, query: Dict[str, str]
    ) -> "DictStore":
        return cls()



import os
import stat
import tempfile
from io import BytesIO
from unittest.mock import Mock
from urllib.parse import (
    quote as url_quote,
    unquote as url_unquote,
    urlparse,
)

import pytest
from basic_store import BasicStore
from conftest import ExtendedKeyspaceTests
from idgens import HashGen, UUIDGen
from url_store import UrlStore

from minimalkv._mixins import ExtendedKeyspaceMixin
from minimalkv.fs import FilesystemStore, WebFilesystemStore


class TestBaseFilesystemStore(BasicStore, UrlStore, UUIDGen, HashGen):
    @pytest.fixture
    def tmpdir(self, tmp_path):
        yield str(tmp_path)

    @pytest.fixture
    def store(self, tmpdir):
        return FilesystemStore(tmpdir)


class TestFilesystemStoreMkdir(TestBaseFilesystemStore):
    def test_concurrent_mkdir(self, tmpdir, mocker):
        # Concurrent instantiation of the store in two threads could lead to
        # the situation where both threads see that the directory does not
        # exists. For one, the call to mkdir succeeds, for the other it fails.
        # This is ok for us as long as the directory exists afterwards.
        makedirs = mocker.patch("os.makedirs")
        makedirs.side_effect = OSError("Failure")
        mocker.patch("os.path.isdir")

        store = FilesystemStore(os.path.join(tmpdir, "test"))
        # We have mocked os.makedirs, so this won't work. But it should
        # pass beyond the OS error and simply fail on writing the file itself.
        with pytest.raises(FileNotFoundError):
            store.put("test", b"test")


class TestFilesystemStoreFileURI(TestBaseFilesystemStore):
    @pytest.mark.skipif(os.name != "posix", reason="Not supported outside posix.")
    def test_correct_file_uri(self, store, tmpdir, key):
        expected = "file://" + tmpdir + "/" + url_quote(key)
        assert store.url_for(key) == expected

    def test_file_uri(self, store, value):
        tmpfile = tempfile.NamedTemporaryFile(delete=False)
        try:
            tmpfile.write(value)
            tmpfile.close()

            key = store.put_file("testkey", tmpfile.name)
            url = store.url_for(key)

            assert url.startswith("file://")
            parts = urlparse(url)

            ndata = open(parts.path, "rb").read()
            assert value == ndata
        finally:
            if os.path.exists(tmpfile.name):
                os.unlink(tmpfile.name)


# runs each test with a nonstandard umask and checks if it is set correctly
class TestFilesystemStoreUmask(TestBaseFilesystemStore):
    @pytest.fixture(scope="class")
    def current_umask(self):
        mask = os.umask(0)
        # re-set umask
        os.umask(mask)
        return mask

    @pytest.fixture(scope="class")
    def perms(self, current_umask):
        # the permissions we expect on files are inverse to the mask
        return 0o666 & (0o777 ^ current_umask)

    def test_file_permission_on_new_file_have_correct_value(self, store, perms, value):
        src = BytesIO(value)

        key = store.put_file("test123", src)

        parts = urlparse(store.url_for(key))
        path = parts.path

        mode = os.stat(path).st_mode
        mask = stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO

        assert mode & mask == perms

    def test_file_permissions_on_moved_in_file_have_correct_value(
        self, store, perms, key, value
    ):
        tmpfile = tempfile.NamedTemporaryFile(delete=False)
        tmpfile.write(value)
        tmpfile.close()
        os.chmod(tmpfile.name, 0o777)
        try:
            key = store.put_file(key, tmpfile.name)

            parts = urlparse(store.url_for(key))
            path = url_unquote(parts.path)

            mode = os.stat(path).st_mode
            mask = stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO

            assert mode & mask == perms
        finally:
            if os.path.exists(tmpfile.name):
                os.unlink(tmpfile.name)


class TestFileStoreSetPermissions(TestFilesystemStoreUmask):
    @pytest.fixture
    def perms(self):
        return 0o612
        self.tmpdir = tempfile.mkdtemp()

    @pytest.fixture
    def store(self, tmpdir, perms):
        return FilesystemStore(tmpdir, perm=perms)


class TestWebFileStore(TestBaseFilesystemStore):
    @pytest.fixture
    def url_prefix(self):
        return "http://some/url/root/"

    @pytest.fixture
    def store(self, tmpdir, url_prefix):
        return WebFilesystemStore(tmpdir, url_prefix)

    def test_url(self, store, url_prefix, key):
        expected = url_prefix + url_quote(key)
        assert store.url_for(key) == expected

    def test_url_callable(self, tmpdir, key):
        prefix = "http://some.prefix.invalid/"
        mock_callable = Mock(return_value=prefix)

        store = WebFilesystemStore(tmpdir, mock_callable)

        expected = prefix + url_quote(key)
        assert store.url_for(key) == expected

        mock_callable.assert_called_with(store, key)


class TestExtendedKeyspaceFilesystemStore(
    TestBaseFilesystemStore, ExtendedKeyspaceTests
):
    @pytest.fixture
    def store(self, tmpdir):
        class ExtendedKeyspaceStore(ExtendedKeyspaceMixin, FilesystemStore):
            pass

        return ExtendedKeyspaceStore(tmpdir)

    def test_prefix_iterator_ossep(self, store, value):
        for k in [
            "a1" + os.sep + "b1",
            "a1" + os.sep + "b1",
            "a2" + os.sep + "b1",
            "a3",
            "a4" + os.sep + "b1" + os.sep + "c1",
            "a4" + os.sep + "b1" + os.sep + "c2",
            "a4" + os.sep + "b2" + os.sep + "c1",
            "a4" + os.sep + "b3",
        ]:
            store.put(k, value)

        out = sorted(store.iter_prefixes(os.sep))
        assert out == [
            "a1" + os.sep,
            "a2" + os.sep,
            "a3",
            "a4" + os.sep,
        ]

        out = sorted(
            store.iter_prefixes(
                os.sep,
                prefix="a4" + os.sep,
            )
        )
        assert out == [
            "a4" + os.sep + "b1" + os.sep,
            "a4" + os.sep + "b2" + os.sep,
            "a4" + os.sep + "b3",
        ]

        out = sorted(
            store.iter_prefixes(
                os.sep,
                prefix="foo" + os.sep,
            )
        )
        assert out == []



import pytest

from minimalkv._get_store import get_store_from_url
from minimalkv._urls import url2dict
from minimalkv.decorator import ReadOnlyDecorator

good_urls = [
    (
        "azure://MYACCOUNT:dead%2Fbeef@1buc-ket1?param1=foo&param2=🍺&eat_more_🍎=1&create_if_missing=true",
        dict(
            type="azure",
            account_name="MYACCOUNT",
            account_key="dead/beef",
            container="1buc-ket1",
            create_if_missing=True,
        ),
    ),
    (
        "azure://MYACCOUNT:deadbeef@1bucket1?param1=foo&param2=🍺&eat_more_🍎=1&create_if_missing=true",
        dict(
            type="azure",
            account_name="MYACCOUNT",
            account_key="deadbeef",
            container="1bucket1",
            create_if_missing=True,
        ),
    ),
    (
        "azure://MYACCOUNT:deadbeef@1bucket1?param1=foo&param2=🍺&eat_more_🍎=&max_connections=5",
        dict(
            type="azure",
            account_name="MYACCOUNT",
            account_key="deadbeef",
            container="1bucket1",
            max_connections=5,
        ),
    ),
    (
        "azure://MYACCOUNT:deadbeef@1bucket1?param1=foo&param2=🍺&eat_more_🍎=&max_connections=5&max_block_size=4194304&&max_single_put_size=67108864",
        dict(
            type="azure",
            account_name="MYACCOUNT",
            account_key="deadbeef",
            container="1bucket1",
            max_connections=5,
            max_block_size=["4194304"],
            max_single_put_size=["67108864"],
        ),
    ),
    ("fs://this/is/a/relative/path", dict(type="fs", path="this/is/a/relative/path")),
    ("fs:///an/absolute/path", dict(type="fs", path="/an/absolute/path")),
    (
        "s3://access_key:secret_key@endpoint:1234/bucketname",
        dict(
            type="s3",
            host="endpoint:1234",
            access_key="access_key",
            secret_key="secret_key",
            bucket="bucketname",
        ),
    ),
    ("redis:///2", dict(type="redis", host="localhost", db=2)),
    ("memory://#wrap:readonly", {"type": "memory", "wrap": "readonly"}),
    ("memory://", dict(type="memory")),
]

bad_urls = [
    (
        "azure://MYACCOUNT:deadb/eef@1buc-ket1?param1=foo&param2=🍺&eat_more_🍎=1&create_if_missing=true",
        ValueError,
    ),
]


def test_raise_on_invalid_store():
    with pytest.raises(ValueError):
        url2dict("dummy://foo/bar")


@pytest.mark.parametrize("url, expected", good_urls)
def test_url2dict(url, expected):
    assert url2dict(url) == expected


@pytest.mark.parametrize("url, raises", bad_urls)
def test_bad_url2dict(url, raises):
    with pytest.raises(raises):
        url2dict(url)


def test_wrapper_scheme():
    assert isinstance(get_store_from_url("memory+readonly://"), ReadOnlyDecorator)


def test_wrapper_fragment():
    assert isinstance(
        get_store_from_url("memory://#wrap:readonly"),
        ReadOnlyDecorator,
    )



import pytest
from bucket_manager import boto3_bucket_reference

from minimalkv._get_store import get_store, get_store_from_url
from minimalkv._urls import url2dict
from minimalkv.net.s3fsstore import S3FSStore

storage = pytest.importorskip("google.cloud.storage")

S3_URL = "s3://minio:miniostorage@127.0.0.1:9000/bucketname?create_if_missing=true&is_secure=false&verify=false"

"""
When using the `s3` scheme in a URL, the new store creation returns an `S3FSStore`.
The old store creation returns a `BotoStore`.
To compare these two implementations, the following tests are run.
"""


def test_new_s3fs_creation():
    expected = S3FSStore(
        bucket=boto3_bucket_reference(
            access_key="minio",
            secret_key="miniostorage",
            host="127.0.0.1",
            port=9000,
            bucket_name="bucketname-minio",
            is_secure=False,
        ),
        verify=False,
    )

    actual = get_store_from_url(S3_URL)
    assert s3fsstores_equal(actual, expected)


def test_equal_access():
    new_store = get_store_from_url(S3_URL)
    old_store = get_store(**url2dict(S3_URL))

    new_store.put("key", b"value")
    assert old_store.get("key") == b"value"


def s3fsstores_equal(store1, store2):
    """Return whether two ``S3FSStore``s are equal.

    The bucket name and other configuration parameters are compared.
    See :func:`from_url` for details on the connection parameters.
    Does NOT compare the credentials or the contents of the bucket!
    """
    return (
        isinstance(store2, S3FSStore)
        and store1.bucket.name == store2.bucket.name
        and store1.bucket.meta.client.meta.endpoint_url
        == store2.bucket.meta.client.meta.endpoint_url
        and store1.object_prefix == store2.object_prefix
        and store1.url_valid_time == store2.url_valid_time
        and store1.reduced_redundancy == store2.reduced_redundancy
        and store1.public == store2.public
        and store1.metadata == store2.metadata
    )



"""In cases where you want to generate IDs automatically, decorators are available.

These should be the outermost decorators, as they change the
signature of some of the put methods slightly.

>>> from minimalkv.memory import DictStore
>>> from minimalkv.idgen import HashDecorator
>>>
>>> store = HashDecorator(DictStore())
>>>
>>> key = store.put(None, b'my_data') #  note the passing of 'None' as key
>>> print(key)
ab0c15b6029fdffce16b393f2d27ca839a76249e
"""

import hashlib
import os
import tempfile
import uuid
from typing import BinaryIO, Optional, Union

from minimalkv.decorator import StoreDecorator


class HashDecorator(StoreDecorator):
    """Hash function decorator.

    Overwrites :meth:`.KeyValueStore.put` and :meth:`.KeyValueStore.put_file`.

    Parameters
    ----------
    decorated_store : KeyValueStore
        Store.
    hashfunc : Callable, optional, default = hashlib.sha1
        Function used for hashing.
    template : str, optional, default = u"{}"
        Template to format hashes.

    """

    def __init__(self, decorated_store, hashfunc=hashlib.sha1, template="{}"):
        self.hashfunc = hashfunc
        self._template = template
        super().__init__(decorated_store)

    def put(self, key: Optional[str], data: bytes, *args, **kwargs):
        """Store bytestring data at key.

        Parameters
        ----------
        key : str or None
            The key under which the data is to be stored. If None, the hash of data is
            used.
        data : bytes
            Data to be stored at key, must be of type  ``bytes``.

        Returns
        -------
        str
            The key under which data was stored.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If storing failed or the file could not be read.
        """
        if key is None:
            key = self._template.format(self.hashfunc(data).hexdigest())

        return self._dstore.put(key, data, *args, **kwargs)  # type: ignore

    def put_file(self, key: Optional[str], file: Union[str, BinaryIO], *args, **kwargs):
        """Store contents of file at key.

        Store data from a file into key. ``file`` can be a string, which will be
        interpreted as a filename, or an object with a ``read()`` method.

        If ``file`` is a filename, the file might be removed while storing to avoid
        unnecessary copies. To prevent this, pass the opened file instead.

        Parameters
        ----------
        key : str or None
            Key where to store data in file. If None, the hash of data is
            used.
        file : BinaryIO or str
            A filename or a file-like object with a read method.

        Returns
        -------
        key: str
            The key under which data was stored.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If there was a problem moving the file in.

        """
        bufsize = 1024 * 1024
        phash = self.hashfunc()

        if key is None:
            if isinstance(file, str):
                with open(file, "rb") as source:
                    while True:
                        buf = source.read(bufsize)
                        phash.update(buf)

                        if len(buf) < bufsize:
                            break

                    return self._dstore.put_file(
                        self._template.format(phash.hexdigest()), file, *args, **kwargs
                    )  # type: ignore
            else:
                tmpfile = tempfile.NamedTemporaryFile(delete=False)
                try:
                    while True:
                        buf = file.read(bufsize)
                        phash.update(buf)
                        tmpfile.write(buf)

                        if len(buf) < bufsize:
                            break

                    tmpfile.close()
                    return self._dstore.put_file(
                        self._template.format(phash.hexdigest()),
                        tmpfile.name,
                        *args,
                        **kwargs,
                    )  # type: ignore
                finally:
                    try:
                        os.unlink(tmpfile.name)
                    except OSError as e:
                        if 2 == e.errno:
                            pass  # file already gone
                        else:
                            raise
        return self._dstore.put_file(key, file, *args, **kwargs)  # type: ignore


class UUIDDecorator(StoreDecorator):
    """UUID generating decorator.

    Overrides :meth:`.KeyValueStore.put` and :meth:`.KeyValueStore.put_file`.
    If key is ``None`` is passed, a new UUID will be generated as the key. The attribute
    ``uuidfunc`` determines which UUID-function to use. 'uuid1'.

    Parameters
    ----------
    store: KeyValueStore
        Store.
    template: str, optional, default = "{}"
        Template to format uuids.

    """

    # There seems to be a bug in the uuid module that prevents initializing
    # `uuidfunc` too early. For that reason, it is a string that will be
    # looked up using :func:`getattr` on the :mod:`uuid` module.
    uuidfunc = "uuid1"

    def __init__(self, store, template="{}"):
        super().__init__(store)
        self._template = template

    def put(self, key: Optional[str], data: bytes, *args, **kwargs) -> str:
        """Store bytestring data at key.

        Parameters
        ----------
        key : str or None
            The key under which the data is to be stored. If None, a uuid is generated.
        data : bytes
            Data to be stored at key, must be of type  ``bytes``.

        Returns
        -------
        str
            The key under which data was stored.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If storing failed or the file could not be read.
        """
        if key is None:
            key = str(getattr(uuid, self.uuidfunc)())

        return self._dstore.put(self._template.format(key), data, *args, **kwargs)  # type: ignore

    def put_file(self, key: Optional[str], file: Union[str, BinaryIO], *args, **kwargs):
        """Store contents of file at key.

        Store data from a file into key. ``file`` can be a string, which will be
        interpreted as a filename, or an object with a ``read()`` method.

        If ``file`` is a filename, the file might be removed while storing to avoid
        unnecessary copies. To prevent this, pass the opened file instead.

        Parameters
        ----------
        key : str or None
            The key under which the data is to be stored. If None, a uuid is generated.
        file : BinaryIO or str
            A filename or a file-like object with a read method.

        Returns
        -------
        key: str
            The key under which data was stored.

        Raises
        ------
        ValueError
            If the key is not valid.
        IOError
            If there was a problem moving the file in.

        """
        if key is None:
            key = str(getattr(uuid, self.uuidfunc)())

        return self._dstore.put_file(self._template.format(key), file, *args, **kwargs)  # type: ignore



import os
import os.path
import shutil
import urllib.parse
from typing import Any, BinaryIO, Callable, Iterator, List, Optional, Union, cast

from minimalkv._key_value_store import KeyValueStore
from minimalkv._mixins import CopyMixin, UrlMixin


class FilesystemStore(KeyValueStore, UrlMixin, CopyMixin):
    """Store data in files on the filesystem under a common directory.

    When files are created, they will receive permissions depending on the current umask
    if ``perm`` is ``None``. Otherwise, permissions are set explicitly.

    Note that when using :func:`put_file` with a filename, an attempt to move the file
    will be made. Permissions and ownership of the file will be preserved that way. If
    ``perm`` is set, permissions will be changed.

    The method :meth:`.url_for` can be used to get a `file://`-URL pointing to the
    internal storage.

    Parameters
    ----------
    root : str
        The base directory for the store.
    perm : int or None, optional, default = None
        The permissions for files in the filesystem store.

    """

    root: str
    perm: Optional[int]
    bufsize: int

    def __init__(self, root: str, perm: Optional[int] = None):
        super().__init__()
        self.root = str(root)
        self.perm = perm
        self.bufsize = 1024 * 1024  # 1m

    def _remove_empty_parents(self, path: str):
        parents = os.path.relpath(path, os.path.abspath(self.root))
        while len(parents) > 0:
            absparent = os.path.join(self.root, parents)
            if os.path.isdir(absparent):
                if len(os.listdir(absparent)) == 0:
                    os.rmdir(absparent)
                else:
                    break
            parents = os.path.dirname(parents)

    def _build_filename(self, key: str) -> str:
        return os.path.abspath(os.path.join(self.root, key))

    def _delete(self, key: str) -> None:
        try:
            targetname = self._build_filename(key)
            os.unlink(targetname)
            self._remove_empty_parents(targetname)
        except OSError as e:
            if not e.errno == 2:
                raise

    def _fix_permissions(self, filename: str) -> None:
        current_umask = os.umask(0)
        os.umask(current_umask)

        perm = self.perm
        if self.perm is None:
            perm = 0o666 & (0o777 ^ current_umask)

        os.chmod(filename, cast(int, perm))

    def _has_key(self, key: str) -> bool:
        return os.path.exists(self._build_filename(key))

    def _open(self, key: str) -> BinaryIO:
        try:
            f = open(self._build_filename(key), "rb")
            return f
        except OSError as e:
            if 2 == e.errno:
                raise KeyError(key) from e
            else:
                raise

    def _copy(self, source: str, dest: str) -> str:
        try:
            source_file_name = self._build_filename(source)
            dest_file_name = self._build_filename(dest)

            self._ensure_dir_exists(os.path.dirname(dest_file_name))
            shutil.copy(source_file_name, dest_file_name)
            self._fix_permissions(dest_file_name)
            return dest
        except OSError as e:
            if 2 == e.errno:
                raise KeyError(source) from e
            else:
                raise

    def _ensure_dir_exists(self, path: str) -> None:
        if not os.path.isdir(path):
            try:
                os.makedirs(path)
            except OSError as e:
                if not os.path.isdir(path):
                    raise e

    def _put_file(self, key: str, file: BinaryIO, *args, **kwargs) -> str:
        bufsize = self.bufsize

        target = self._build_filename(key)
        self._ensure_dir_exists(os.path.dirname(target))

        with open(target, "wb") as f:
            while True:
                buf = file.read(bufsize)
                f.write(buf)
                if len(buf) < bufsize:
                    break

        # when using umask, correct permissions are automatically applied
        # only chmod is necessary
        if self.perm is not None:
            self._fix_permissions(target)

        return key

    def _put_filename(self, key: str, filename: str, *args, **kwargs) -> str:
        target = self._build_filename(key)
        self._ensure_dir_exists(os.path.dirname(target))
        shutil.move(filename, target)

        # we do not know the permissions of the source file, rectify
        self._fix_permissions(target)
        return key

    def _url_for(self, key: str) -> str:
        full = os.path.abspath(self._build_filename(key))
        parts = full.split(os.sep)
        location = "/".join(urllib.parse.quote(p, safe="") for p in parts)
        return "file://" + location

    def keys(self, prefix: str = "") -> List[str]:
        """List all keys in the store starting with prefix.

        Parameters
        ----------
        prefix : str, optional, default = ''
            Only list keys starting with prefix. List all keys if empty.

        """
        root = os.path.abspath(self.root)
        result = []
        for dp, _, fn in os.walk(root):
            for f in fn:
                key = os.path.join(dp, f)[len(root) + 1 :]
                if key.startswith(prefix):
                    result.append(key)
        return result

    def iter_keys(self, prefix: str = "") -> Iterator[str]:
        """Iterate over all keys in the store starting with prefix.

        Parameters
        ----------
        prefix : str, optional, default = ''
            Only iterate over keys starting with prefix. Iterate over all keys if empty.

        """
        return iter(self.keys(prefix))

    def iter_prefixes(self, delimiter: str, prefix: str = "") -> Iterator[str]:
        """Iterate over unique prefixes in the store up to delimiter, starting with prefix.

        If ``prefix`` contains ``delimiter``, return the prefix up to the first
        occurence of delimiter after the prefix.

        Parameters
        ----------
        delimiter : str, optional, default = ''
            Delimiter up to which to iterate over prefixes.
        prefix : str, optional, default = ''
            Only iterate over prefixes starting with prefix.

        """
        if delimiter != os.sep:
            return super().iter_prefixes(
                delimiter,
                prefix,
            )
        return self._iter_prefixes_efficient(delimiter, prefix)

    def _iter_prefixes_efficient(
        self, delimiter: str, prefix: str = ""
    ) -> Iterator[str]:
        if delimiter in prefix:
            pos = prefix.rfind(delimiter)
            search_prefix: Optional[str] = prefix[:pos]
            path = os.path.join(self.root, cast(str, search_prefix))
        else:
            search_prefix = None
            path = self.root

        try:
            for k in os.listdir(path):
                subpath = os.path.join(path, k)

                if search_prefix is not None:
                    k = os.path.join(search_prefix, k)

                if os.path.isdir(subpath):
                    k += delimiter

                if k.startswith(prefix):
                    yield k
        except OSError:
            # path does not exists
            pass


class WebFilesystemStore(FilesystemStore):
    """FilesystemStore supporting generating URLS for web applications.

    The most common use is to make the ``root`` directory of the filesystem store
    available through a webserver.

    Note that the prefix is simply prepended to the relative URL for the key. It must
    therefore include a trailing slash in most cases.

    ``url_prefix`` may also be a callable, in which case it gets called with the
    filestore and key as an argument and should return an url_prefix.

    Parameters
    ----------
    root : str
        The base directory for the store.
    perm : int or None, optional, default = None
        The permissions for files in the filesystem store.
    url_prefix : Callable or str
        Prefix that will get prepended to every url.

    Example
    -------
    >>> from minimalkv.fs import WebFilesystemStore
    >>> webserver_url_prefix = 'https://some.domain.invalid/files/'
    >>> webserver_root = '/var/www/some.domain.invalid/www-data/files/'
    >>> store = WebFilesystemStore(webserver_root, webserver_url_prefix)
    >>> print(store.url_for(u'some_key'))
    https://some.domain.invalid/files/some_key

    >>> from minimalkv.fs import WebFilesystemStore
    >>> webserver_url_prefix = 'https://some.domain.invalid/files/'
    >>> webserver_root = '/var/www/some.domain.invalid/www-data/files/'
    >>> prefix_func = lambda store, key: webserver_url_prefix
    >>> store = WebFilesystemStore(webserver_root, prefix_func)
    >>> print(store.url_for(u'some_key'))
    https://some.domain.invalid/files/some_key
    """

    def __init__(
        self, root, url_prefix: Union[Callable[[Any, str], str], str], **kwargs
    ):
        super().__init__(root, **kwargs)

        self.url_prefix = url_prefix

    def _url_for(self, key: str) -> str:
        rel = key

        if callable(self.url_prefix):
            stem: str = self.url_prefix(self, key)
        else:
            stem = self.url_prefix
        return stem + urllib.parse.quote(rel, safe="")
